#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
Constants, state and parsing primitives for Primary Drawing Orders.
"""
from io import BytesIO

from pyrdp.enum.orders import DrawingOrderControlFlags as ControlFlags
from pyrdp.core.packing import Uint8, Int8, Int16LE, Uint16LE, Uint32LE
from .common import GlyphV2, Bounds
from .secondary import BMF_BPP, CACHED_BRUSH

# This follows the PrimaryDrawOrderType enum.
ORDERTYPE_FIELDBYTES = [1, 2, 1, 0, 0, 0, 0, 1, 1, 2, 1, 1, 0, 2, 3, 1, 2, 2, 2, 2, 1, 2, 1, 0, 2, 1, 2, 3]

BACKMODE_TRANSPARENT = 0x01
BACKMODE_OPAQUE = 0x02


def read_field_flags(s: BytesIO, flags: int, orderType: int) -> int:
    """Reads encoded field flags."""

    # REFACTOR: This could be internal to the context class.
    assert orderType >= 0 and orderType < len(ORDERTYPE_FIELDBYTES)

    fieldBytes = ORDERTYPE_FIELDBYTES[orderType]
    assert fieldBytes != 0  # Should be a valid orderType

    if flags & ControlFlags.TS_ZERO_FIELD_BYTE_BIT0:
        fieldBytes -= 1

    if flags & ControlFlags.TS_ZERO_FIELD_BYTE_BIT1:
        if fieldBytes > 1:
            fieldBytes -= 2
        else:
            fieldBytes = 0

    fieldFlags = 0
    for i in range(fieldBytes):
        fieldFlags |= Uint8.unpack(s) << (i * 8)

    return fieldFlags


def read_coord(s: BytesIO, delta: bool, prev: int):
    if delta:
        return prev + Int8.unpack(s)
    else:
        return Int16LE.unpack(s)


def read_delta(s: BytesIO) -> int:
    msb = Uint8.unpack(s)
    val = msb | ~0x3F if msb & 0x40 else msb & 0x3F
    if msb & 0x80:
        val = (val << 8) | Uint8.unpack(s)
    return val


def read_rgb(s: BytesIO) -> int:
    """Read an RGB color encoded as 0xBBGGRR."""
    r = Uint8.unpack(s)
    g = Uint8.unpack(s)
    b = Uint8.unpack(s)
    return r | g << 8 | b << 16


def read_delta_points(s: BytesIO, n: int, x0: int, y0: int) -> [(int, int)]:
    """
    Read an array of delta encoded points.

    :param s: The data stream to parse the delta points from.
    :param n: The number of points that are encoded in the stream.
    :param x0: The initial value of x.
    :param y0: The initial value of y.

    A points is represented as an (x,y)-tuple.

    This function converts the deltas into absolute coordinates.

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpegdi/6c7b2a52-103c-4a7d-a2a9-997416d4a475
    """

    zeroBitsLen = ((n + 3) // 4)
    zeroBits = s.read(zeroBitsLen)

    dx = x0
    dy = y0

    points = []
    for i in range(n):

        # Next zeroBits byte.
        if i % 4 == 0:
            flags = zeroBits[i // 4]

        x = (read_delta(s) + dx) if not flags & 0x80 else dx
        y = (read_delta(s) + dy) if not flags & 0x40 else dy
        flags <<= 2
        points.append((x, y))

        # Update previous point coords.
        dx = x
        dy = y

    return points


def read_delta_rectangles(s: BytesIO, n: int) -> [(int, int, int, int)]:
    """
    Read an array of delta encoded rectangles.

    :param s: The data stream to parse the rectangles from.
    :param n: The number of rectangles encoded in the stream.

    A rectangles is represented as a (left, top, width, height)-tuple.

    This function converts the deltas into absolute coordinates.

    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpegdi/b89f2058-b180-4da0-9bd1-aa694c87768c
    """

    zeroBitsSize = (n + 1) // 2
    zeroBits = s.read(zeroBitsSize)

    rectangles = []
    dl = dt = dw = dh = 0

    for i in range(n):
        if i % 2 == 0:
            flags = zeroBits[i // 2]

        left = read_delta(s) + dl if not flags & 0x80 else dl
        top = read_delta(s) + dt if not flags & 0x40 else dt
        width = read_delta(s) if not flags & 0x20 else dw
        height = read_delta(s) if not flags & 0x10 else dh
        flags <<= 4
        rectangles.append((left, top, width, height))

        # Update previous rectangle coords.
        dl = left
        dt = top
        dw = width
        dh = height
    return rectangles


class PrimaryContext:
    """Primary drawing order context."""

    def __init__(self):
        # The field flags for the current order.
        self.fieldFlags: int = 0

        # Whether coordinates are being sent as a delta.
        self.deltaCoords: bool = False

        # A cache of the previous order type received.
        self.orderType: int = None

        # The configured bounding rectangle
        self.bounds: Bounds = Bounds()
        self.bounded: bool = False

        # Track state for each drawing order.
        self.dstBlt = DstBlt(self)
        self.patBlt = PatBlt(self)
        self.scrBlt = ScrBlt(self)
        self.drawNineGrid = DrawNineGrid(self)
        self.multiDrawNineGrid = MultiDrawNineGrid(self)
        self.lineTo = LineTo(self)
        self.opaqueRect = OpaqueRect(self)
        self.saveBitmap = SaveBitmap(self)
        self.memBlt = MemBlt(self)
        self.mem3Blt = Mem3Blt(self)
        self.multiDstBlt = MultiDstBlt(self)
        self.multiPatBlt = MultiPatBlt(self)
        self.multiScrBlt = MultiScrBlt(self)
        self.multiOpaqueRect = MultiOpaqueRect(self)
        self.fastIndex = FastIndex(self)
        self.polygonSc = PolygonSc(self)
        self.polygonCb = PolygonCb(self)
        self.polyLine = PolyLine(self)
        self.fastGlyph = FastGlyph(self)
        self.ellipseSc = EllipseSc(self)
        self.ellipseCb = EllipseCb(self)
        self.glyphIndex = GlyphIndex(self)

    def update(self, s: BytesIO, flags: int):
        """
        Update the context when parsing a new primary order.

        This method should be called at the beginning of every new
        primary order to process contextual changes.

        :param s BytesIO: The raw byte stream
        :param flags int: The controlFlags received in the UPDATE PDU.

        :return: The orderType to act upon.
        """

        if flags & ControlFlags.TS_TYPE_CHANGE:
            self.orderType = Uint8.unpack(s)
        assert self.orderType is not None

        self.fieldFlags = read_field_flags(s, flags, self.orderType)

        # Process bounding rectangle updates
        if flags & ControlFlags.TS_BOUNDS:
            self.bounded = True
            if not flags & ControlFlags.TS_ZERO_BOUNDS_DELTAS:
                self.bounds.update(s)
        else:
            self.bounded = False

        self.deltaCoords = flags & ControlFlags.TS_DELTA_COORDS != 0

        return self.orderType

    def field(self, n: int):
        """Check whether field `n` is present in the message."""
        return self.fieldFlags & (1 << (n - 1)) != 0


class Brush:
    def __init__(self):
        self.x = self.y = 0
        self.style = 0
        self.hatch = 0
        self.data = None
        self.index = None
        self.bpp = 0

    def update(self, s: BytesIO, flags: int):
        if flags & 0b00001:
            self.x = Uint8.unpack(s)
        if flags & 0b00010:
            self.y = Uint8.unpack(s)
        if flags & 0b00100:
            self.style = Uint8.unpack(s)
        if flags & 0b01000:
            self.hatch = Uint8.unpack(s)
        if flags & 0b10000:
            self.data = (s.read(7) + bytes([self.hatch]))[::-1]

        if self.style & CACHED_BRUSH:
            self.index = self.hatch
            self.bpp = BMF_BPP[self.style & 0x07]
            if self.bpp == 0:
                self.bpp = 1

        return self


class DstBlt:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx
        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0
        self.rop = 0

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.x = read_coord(s, self.ctx.deltaCoords, self.x)
        if self.ctx.field(2):
            self.y = read_coord(s, self.ctx.deltaCoords, self.y)
        if self.ctx.field(3):
            self.w = read_coord(s, self.ctx.deltaCoords, self.w)
        if self.ctx.field(4):
            self.h = read_coord(s, self.ctx.deltaCoords, self.h)
        if self.ctx.field(5):
            self.rop = Uint8.unpack(s)

        return self

    def __str__(self):
        return f'<DstBlt ({self.x}, {self.y}) {self.w}x{self.h} Rop={self.rop}>'


class PatBlt:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx
        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0
        self.rop = 0
        self.bg = 0
        self.fg = 0
        self.brush = Brush()

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.x = read_coord(s, self.ctx.deltaCoords, self.x)
        if self.ctx.field(2):
            self.y = read_coord(s, self.ctx.deltaCoords, self.y)
        if self.ctx.field(3):
            self.w = read_coord(s, self.ctx.deltaCoords, self.w)
        if self.ctx.field(4):
            self.h = read_coord(s, self.ctx.deltaCoords, self.h)
        if self.ctx.field(5):
            self.rop = Uint8.unpack(s)
        if self.ctx.field(6):
            self.bg = read_rgb(s)
        if self.ctx.field(7):
            self.fg = read_rgb(s)

        self.brush.update(s, self.ctx.fieldFlags >> 7)

        return self

    def __str__(self):
        return f'<PatBlt ({self.x}, {self.y}) {self.w}x{self.h} Rop={self.rop}>'


class ScrBlt:
    """
    2.2.2.2.1.1.2.7
    """

    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.nLeftRect = 0
        self.nTopRect = 0
        self.nWidth = 0
        self.nHeight = 0
        self.bRop = 0
        self.nXSrc = 0
        self.nYSrc = 0

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.nLeftRect = read_coord(s, self.ctx.deltaCoords, self.nLeftRect)
        if self.ctx.field(2):
            self.nTopRect = read_coord(s, self.ctx.deltaCoords, self.nTopRect)
        if self.ctx.field(3):
            self.nWidth = read_coord(s, self.ctx.deltaCoords, self.nWidth)
        if self.ctx.field(4):
            self.nHeight = read_coord(s, self.ctx.deltaCoords, self.nHeight)
        if self.ctx.field(5):
            self.bRop = Uint8.unpack(s)
        if self.ctx.field(6):
            self.nXSrc = read_coord(s, self.ctx.deltaCoords, self.nXSrc)
        if self.ctx.field(7):
            self.nYSrc = read_coord(s, self.ctx.deltaCoords, self.nYSrc)

        return self

    def __str__(self):
        return (f'<ScrBlt Src=({self.nXSrc},{self.nYSrc}) OP={self.bRop}'
                f' Dst=({self.nLeftRect}, {self.nTopRect}, {self.nWidth}, {self.nHeight})')


class DrawNineGrid:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.srcLeft = 0
        self.srcTop = 0
        self.srcRight = 0
        self.srcBottom = 0
        self.bitmapId = 0

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.srcLeft = read_coord(s, self.ctx.deltaCoords, self.srcLeft)
        if self.ctx.field(2):
            self.srcTop = read_coord(s, self.ctx.deltaCoords, self.srcTop)
        if self.ctx.field(3):
            self.srcRight = read_coord(s, self.ctx.deltaCoords, self.srcRight)
        if self.ctx.field(4):
            self.srcBottom = read_coord(s, self.ctx.deltaCoords, self.srcBottom)
        if self.ctx.field(5):
            self.bitmapId = Uint16LE.unpack(s)

        return self

    def __str__(self):
        return '<DrawNineGrid>'


class MultiDrawNineGrid:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.srcLeft = 0
        self.srcTop = 0
        self.srcRight = 0
        self.srcBottom = 0
        self.bitmapId = 0
        self.nDeltaEntries = 0
        self.cbData = 0
        self.rectangles = []

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.srcLeft = read_coord(s, self.ctx.deltaCoords, self.srcLeft)
        if self.ctx.field(2):
            self.srcTop = read_coord(s, self.ctx.deltaCoords, self.srcTop)
        if self.ctx.field(3):
            self.srcRight = read_coord(s, self.ctx.deltaCoords, self.srcRight)
        if self.ctx.field(4):
            self.srcBottom = read_coord(s, self.ctx.deltaCoords, self.srcBottom)
        if self.ctx.field(5):
            self.bitmapId = Uint16LE.unpack(s)
        if self.ctx.field(6):
            self.nDeltaEntries = Uint8.unpack(s)

        if self.ctx.field(7):
            self.cbData = Uint16LE.unpack(s)
            self.rectangles = read_delta_rectangles(s, self.nDeltaEntries)

        return self

    def __str__(self):
        return '<MultiDrawNineGrid>'


class LineTo:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.bgMode = 0
        self.x0 = 0
        self.y0 = 0
        self.x1 = 0
        self.y1 = 0
        self.bg = 0
        self.rop2 = 0
        self.penStyle = 0
        self.penWidth = 0
        self.penColor = 0

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.bgMode = Uint16LE.unpack(s)
        if self.ctx.field(2):
            self.x0 = read_coord(s, self.ctx.deltaCoords, self.x0)
        if self.ctx.field(3):
            self.y0 = read_coord(s, self.ctx.deltaCoords, self.y0)
        if self.ctx.field(4):
            self.x1 = read_coord(s, self.ctx.deltaCoords, self.x1)
        if self.ctx.field(5):
            self.y1 = read_coord(s, self.ctx.deltaCoords, self.y1)
        if self.ctx.field(6):
            self.bg = read_rgb(s)
        if self.ctx.field(7):
            self.rop2 = Uint8.unpack(s)
        if self.ctx.field(8):
            self.penStyle = Uint8.unpack(s)
        if self.ctx.field(9):
            self.penWidth = Uint8.unpack(s)
        if self.ctx.field(10):
            self.penColor = read_rgb(s)

        return self

    def __str__(self):
        return '<LineTo>'


class OpaqueRect:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0
        self.color = 0  # 0xBBGGRR

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.x = read_coord(s, self.ctx.deltaCoords, self.x)
        if self.ctx.field(2):
            self.y = read_coord(s, self.ctx.deltaCoords, self.y)
        if self.ctx.field(3):
            self.w = read_coord(s, self.ctx.deltaCoords, self.w)
        if self.ctx.field(4):
            self.h = read_coord(s, self.ctx.deltaCoords, self.h)

        if self.ctx.field(5):
            r = Uint8.unpack(s)
            self.color = (self.color & 0x00FFFF00) | r
        if self.ctx.field(6):
            g = Uint8.unpack(s)
            self.color = (self.color & 0x00FF00FF) | (g << 8)
        if self.ctx.field(7):
            b = Uint8.unpack(s)
            self.color = (self.color & 0x0000FFFF) | (b << 16)

        return self

    def __str__(self):
        return f'<OpaqueRect ({self.x}, {self.y}) {self.w}x{self.h}) Color={self.color:06X}>'


class SaveBitmap:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.savedBitmapPosition = 0
        self.nLeftRect = 0
        self.nTopRect = 0
        self.nRightRect = 0
        self.nBottomRect = 0
        self.operation = 0

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.savedBitmapPosition = Uint32LE.unpack(s)
        if self.ctx.field(2):
            self.nLeftRect = read_coord(s, self.ctx.deltaCoords, self.nLeftRect)
        if self.ctx.field(3):
            self.nTopRect = read_coord(s, self.ctx.deltaCoords, self.nTopRect)
        if self.ctx.field(4):
            self.nRightRect = read_coord(s, self.ctx.deltaCoords, self.nRightRect)
        if self.ctx.field(5):
            self.nBottomRect = read_coord(s, self.ctx.deltaCoords, self.nBottomRect)
        if self.ctx.field(6):
            self.operation = Uint8.unpack(s)

        return self

    def __str__(self):
        return '<SaveBitmap>'


class MemBlt:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        # Blit rectangle.
        self.left = self.top = self.width = self.height = 0

        # Source buffer offsets.
        self.xSrc = self.ySrc = 0

        self.cacheIndex = 0
        self.cacheId = 0
        self.colorIndex = 0

    def update(self, s: BytesIO) -> 'MemBlt':
        ctx = self.ctx

        if ctx.field(1):
            self.cacheId = Uint16LE.unpack(s)
        if ctx.field(2):
            self.left = read_coord(s, ctx.deltaCoords, self.left)
        if ctx.field(3):
            self.top = read_coord(s, ctx.deltaCoords, self.top)
        if ctx.field(4):
            self.width = read_coord(s, ctx.deltaCoords, self.width)
        if ctx.field(5):
            self.height = read_coord(s, ctx.deltaCoords, self.height)
        if ctx.field(6):
            self.rop = Uint8.unpack(s)
        if ctx.field(7):
            self.xSrc = read_coord(s, ctx.deltaCoords, self.xSrc)
        if ctx.field(8):
            self.ySrc = read_coord(s, ctx.deltaCoords, self.ySrc)
        if ctx.field(9):
            self.cacheIndex = Uint16LE.unpack(s)

        self.colorIndex = self.cacheId >> 8
        self.cacheId = self.cacheId & 0xFF

        return self

    def __str__(self):
        return (f'<MemBlt ({self.xSrc},{self.ySrc}) OP={self.rop} L={self.left} T={self.top}'
                f' W={self.width} H={self.height} cacheIndex={self.cacheIndex} cacheId={self.cacheId}'
                f' colorIdx={self.colorIndex}>')


class Mem3Blt:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx
        self.brush = Brush()

        self.cacheId = 0
        self.left = 0
        self.top = 0
        self.width = 0
        self.height = 0
        self.rop = 0
        self.nXSrc = 0
        self.nYSrc = 0
        self.bg = 0
        self.fg = 0
        self.cacheIndex = 0
        self.colorIndex = 0
        self.cacheId = 0

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.cacheId = Uint16LE.unpack(s)
        if self.ctx.field(2):
            self.left = read_coord(s, self.ctx.deltaCoords, self.left)
        if self.ctx.field(3):
            self.top = read_coord(s, self.ctx.deltaCoords, self.top)
        if self.ctx.field(4):
            self.width = read_coord(s, self.ctx.deltaCoords, self.width)
        if self.ctx.field(5):
            self.height = read_coord(s, self.ctx.deltaCoords, self.height)
        if self.ctx.field(6):
            self.rop = Uint8.unpack(s)
        if self.ctx.field(7):
            self.nXSrc = read_coord(s, self.ctx.deltaCoords, self.nXSrc)
        if self.ctx.field(8):
            self.nYSrc = read_coord(s, self.ctx.deltaCoords, self.nYSrc)
        if self.ctx.field(9):
            self.bg = read_rgb(s)
        if self.ctx.field(10):
            self.fg = read_rgb(s)

        self.brush.update(s, self.ctx.fieldFlags >> 10)

        if self.ctx.field(16):
            self.cacheIndex = Uint16LE.unpack(s)

        self.colorIndex = self.cacheId >> 8
        self.cacheId = self.cacheId & 0xFF

        return self

    def __str__(self):
        return '<Mem3Blt>'


class MultiDstBlt:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0
        self.rop = 0
        self.numRectangles = 0
        self.cbData = 0
        self.rectangles = []

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.x = read_coord(s, self.ctx.deltaCoords, self.x)
        if self.ctx.field(2):
            self.y = read_coord(s, self.ctx.deltaCoords, self.y)
        if self.ctx.field(3):
            self.w = read_coord(s, self.ctx.deltaCoords, self.w)
        if self.ctx.field(4):
            self.h = read_coord(s, self.ctx.deltaCoords, self.h)
        if self.ctx.field(5):
            self.rop = Uint8.unpack(s)
        if self.ctx.field(6):
            self.numRectangles = Uint8.unpack(s)

        if self.ctx.field(7):
            self.cbData = Uint16LE.unpack(s)
            self.rectangles = read_delta_rectangles(s, self.numRectangles)

        return self

    def __str__(self):
        return '<MultiDstBlt>'


class MultiPatBlt:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx
        self.brush = Brush()

        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0
        self.rop = 0
        self.bg = 0
        self.fg = 0
        self.numRectangles = 0
        self.cbData = 0
        self.rectangles = []

    def update(self, s: BytesIO):

        if self.ctx.field(1):
            self.x = read_coord(s, self.ctx.deltaCoords, self.x)
        if self.ctx.field(2):
            self.y = read_coord(s, self.ctx.deltaCoords, self.y)
        if self.ctx.field(3):
            self.w = read_coord(s, self.ctx.deltaCoords, self.w)
        if self.ctx.field(4):
            self.h = read_coord(s, self.ctx.deltaCoords, self.h)
        if self.ctx.field(5):
            self.rop = Uint8.unpack(s)
        if self.ctx.field(6):
            self.bg = read_rgb(s)
        if self.ctx.field(7):
            self.fg = read_rgb(s)

        self.brush.update(s, self.ctx.fieldFlags >> 7)

        if self.ctx.field(13):
            self.numRectangles = Uint8.unpack(s)

        if self.ctx.field(14):
            self.cbData = Uint16LE.unpack(s)
            self.rectangles = read_delta_rectangles(s, self.numRectangles)

        return self

    def __str__(self):
        return '<MultiPatBlt>'


class MultiScrBlt:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.nLeftRect = 0
        self.nTopRect = 0
        self.nWidth = 0
        self.nHeight = 0
        self.bRop = 0
        self.nXSrc = 0
        self.nYSrc = 0
        self.numRectangles = 0
        self.cbData = 0
        self.rectangles = []

    def update(self, s: BytesIO):

        if self.ctx.field(1):
            self.nLeftRect = read_coord(s, self.ctx.deltaCoords, self.nLeftRect)
        if self.ctx.field(2):
            self.nTopRect = read_coord(s, self.ctx.deltaCoords, self.nTopRect)
        if self.ctx.field(3):
            self.nWidth = read_coord(s, self.ctx.deltaCoords, self.nWidth)
        if self.ctx.field(4):
            self.nHeight = read_coord(s, self.ctx.deltaCoords, self.nHeight)
        if self.ctx.field(5):
            self.bRop = Uint8.unpack(s)
        if self.ctx.field(6):
            self.nXSrc = read_coord(s, self.ctx.deltaCoords, self.nXSrc)
        if self.ctx.field(7):
            self.nYSrc = read_coord(s, self.ctx.deltaCoords, self.nYSrc)
        if self.ctx.field(8):
            self.numRectangles = Uint8.unpack(s)

        if self.ctx.field(9):
            self.cbData = Uint16LE.unpack(s)
            self.rectangles = read_delta_rectangles(s, self.numRectangles)

        return self

    def __str__(self):
        return '<MultiScrBlt>'


class MultiOpaqueRect:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.nLeftRect = 0
        self.nTopRect = 0
        self.nWidth = 0
        self.nHeight = 0
        self.color = 0  # 0xBBGGRR
        self.numRectangles = 0
        self.cbData = 0
        self.rectangles = []

    def update(self, s: BytesIO):

        if self.ctx.field(1):
            self.nLeftRect = read_coord(s, self.ctx.deltaCoords, self.nLeftRect)
        if self.ctx.field(2):
            self.nTopRect = read_coord(s, self.ctx.deltaCoords, self.nTopRect)
        if self.ctx.field(3):
            self.nWidth = read_coord(s, self.ctx.deltaCoords, self.nWidth)
        if self.ctx.field(4):
            self.nHeight = read_coord(s, self.ctx.deltaCoords, self.nHeight)

        if self.ctx.field(5):
            r = Uint8.unpack(s)
            self.color = (self.color & 0x00FFFF00) | r
        if self.ctx.field(6):
            g = Uint8.unpack(s)
            self.color = (self.color & 0x00FF00FF) | (g << 8)
        if self.ctx.field(7):
            b = Uint8.unpack(s)
            self.color = (self.color & 0x0000FFFF) | (b << 16)

        if self.ctx.field(8):
            self.numRectangles = Uint8.unpack(s)

        if self.ctx.field(9):
            self.cbData = Uint16LE.unpack(s)
            self.rectangles = read_delta_rectangles(s, self.numRectangles)

        return self

    def __str__(self):
        return f'<MultiOpaqueRect Color=#{self.color:06X}>'


class FastIndex:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.cacheId = 0
        self.ulCharInc = 0
        self.flAccel = 0
        self.bg = 0
        self.fg = 0
        self.bkLeft = 0
        self.bkTop = 0
        self.bkRight = 0
        self.bkBottom = 0
        self.opLeft = 0
        self.opTop = 0
        self.opRight = 0
        self.opBottom = 0
        self.x = 0
        self.y = 0

        self.data = b''

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.cacheId = Uint8.unpack(s)
        if self.ctx.field(2):
            self.ulCharInc = Uint8.unpack(s)
            self.flAccel = Uint8.unpack(s)
        if self.ctx.field(3):
            self.bg = read_rgb(s)
        if self.ctx.field(4):
            self.fg = read_rgb(s)
        if self.ctx.field(5):
            self.bkLeft = read_coord(s, self.ctx.deltaCoords, self.bkLeft)
        if self.ctx.field(6):
            self.bkTop = read_coord(s, self.ctx.deltaCoords, self.bkTop)
        if self.ctx.field(7):
            self.bkRight = read_coord(s, self.ctx.deltaCoords, self.bkRight)
        if self.ctx.field(8):
            self.bkBottom = read_coord(s, self.ctx.deltaCoords, self.bkBottom)
        if self.ctx.field(9):
            self.opLeft = read_coord(s, self.ctx.deltaCoords, self.opLeft)
        if self.ctx.field(10):
            self.opTop = read_coord(s, self.ctx.deltaCoords, self.opTop)
        if self.ctx.field(11):
            self.opRight = read_coord(s, self.ctx.deltaCoords, self.opRight)
        if self.ctx.field(12):
            self.opBottom = read_coord(s, self.ctx.deltaCoords, self.opBottom)
        if self.ctx.field(13):
            self.x = read_coord(s, self.ctx.deltaCoords, self.x)
        if self.ctx.field(14):
            self.y = read_coord(s, self.ctx.deltaCoords, self.y)

        if self.ctx.field(15):
            cbData = Uint8.unpack(s)
            self.data = s.read(cbData)

        return self

    def __str__(self):
        return f'<FastIndex ({self.x}, {self.y}) Len={len(self.data)}>'


class PolygonSc:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.x0 = 0
        self.y0 = 0
        self.rop2 = 0
        self.fillMode = 0
        self.brushColor = 0
        self.cbData = 0
        self.numPoints = 0
        self.points = []

    def update(self, s: BytesIO):
        num = self.numPoints

        if self.ctx.field(1):
            self.x0 = read_coord(s, self.ctx.deltaCoords, self.x0)
        if self.ctx.field(2):
            self.y0 = read_coord(s, self.ctx.deltaCoords, self.y0)
        if self.ctx.field(3):
            self.rop2 = Uint8.unpack(s)
        if self.ctx.field(4):
            self.fillMode = Uint8.unpack(s)
        if self.ctx.field(5):
            self.brushColor = read_rgb(s)

        if self.ctx.field(6):
            num = Uint8.unpack(s)

        if self.ctx.field(7):
            self.cbData = Uint8.unpack(s)
            self.numPoints = num
            self.points = read_delta_points(s, self.numPoints, self.x0, self.y0)

        return self

    def __str__(self):
        return '<PolygonSc>'


class PolygonCb:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx
        self.brush = Brush()

        self.x0 = 0
        self.y0 = 0
        self.rop2 = 0
        self.fillMode = 0
        self.bg = 0
        self.fg = 0
        self.cbData = 0
        self.numPoints = 0
        self.points = []
        self.bgMode = BACKMODE_OPAQUE

    def update(self, s: BytesIO):

        num = self.numPoints
        if self.ctx.field(1):
            self.x0 = read_coord(s, self.ctx.deltaCoords, self.x0)
        if self.ctx.field(2):
            self.y0 = read_coord(s, self.ctx.deltaCoords, self.y0)
        if self.ctx.field(3):
            self.rop2 = Uint8.unpack(s)
        if self.ctx.field(4):
            self.fillMode = Uint8.unpack(s)
        if self.ctx.field(5):
            self.bg = read_rgb(s)
        if self.ctx.field(6):
            self.fg = read_rgb(s)

        self.brush.update(s, self.ctx.fieldFlags >> 6)

        if self.ctx.field(12):
            num = Uint8.unpack(s)

        if self.ctx.field(13):
            self.cbData = Uint8.unpack(s)
            self.numPoints = num
            self.points = read_delta_points(s, self.numPoints, self.x0, self.y0)

        self.bgMode = BACKMODE_TRANSPARENT if self.rop2 & 0x80 else BACKMODE_OPAQUE
        self.rop2 = self.rop2 & 0x1F

        return self

    def __str__(self):
        return '<PolygonCb>'


class PolyLine:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.x0 = 0
        self.y0 = 0
        self.rop2 = 0
        self.penColor = 0
        self.cbData = 0
        self.numPoints = 0
        self.points = []

    def update(self, s: BytesIO):
        num = self.numPoints
        if self.ctx.field(1):
            self.x0 = read_coord(s, self.ctx.deltaCoords, self.x0)
        if self.ctx.field(2):
            self.y0 = read_coord(s, self.ctx.deltaCoords, self.y0)
        if self.ctx.field(3):
            self.rop2 = Uint8.unpack(s)
        if self.ctx.field(4):
            s.read(2)  # unused (brushCacheIndex)
        if self.ctx.field(5):
            self.penColor = read_rgb(s)
        if self.ctx.field(6):
            num = Uint8.unpack(s)

        if self.ctx.field(7):
            self.cbData = Uint8.unpack(s)
            self.numPoints = num
            self.points = read_delta_points(s, self.numPoints, self.x0, self.y0)

        return self

    def __str__(self):
        return '<PolyLine>'


class FastGlyph:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx
        self.cacheId = 0
        self.cacheIndex = 0
        self.glyph = None
        self.ulCharInc = 0
        self.flAccel = 0
        self.bg = 0
        self.fg = 0

        # Text background coords.
        self.bkLeft = 0
        self.bkTop = 0
        self.bkRight = 0
        self.bkBottom = 0

        # Opaque rectangle coords. (0 -> same as Bk*)
        self.opLeft = 0
        self.opTop = 0
        self.opRight = 0
        self.opBottom = 0

        # Position of the glyph.
        self.x = 0
        self.y = 0

    def update(self, s: BytesIO):
        if self.ctx.field(1):
            self.cacheId = Uint8.unpack(s)
        if self.ctx.field(2):
            self.ulCharInc = Uint8.unpack(s)
            self.flAccel = Uint8.unpack(s)
        if self.ctx.field(3):
            self.bg = read_rgb(s)
        if self.ctx.field(4):
            self.fg = read_rgb(s)
        if self.ctx.field(5):
            self.bkLeft = read_coord(s, self.ctx.deltaCoords, self.bkLeft)
        if self.ctx.field(6):
            self.bkTop = read_coord(s, self.ctx.deltaCoords, self.bkTop)
        if self.ctx.field(7):
            self.bkRight = read_coord(s, self.ctx.deltaCoords, self.bkRight)
        if self.ctx.field(8):
            self.bkBottom = read_coord(s, self.ctx.deltaCoords, self.bkBottom)
        if self.ctx.field(9):
            self.opLeft = read_coord(s, self.ctx.deltaCoords, self.opLeft)
        if self.ctx.field(10):
            self.opTop = read_coord(s, self.ctx.deltaCoords, self.opTop)
        if self.ctx.field(11):
            self.opRight = read_coord(s, self.ctx.deltaCoords, self.opRight)
        if self.ctx.field(12):
            self.opBottom = read_coord(s, self.ctx.deltaCoords, self.opBottom)
        if self.ctx.field(13):
            self.x = read_coord(s, self.ctx.deltaCoords, self.x)
        if self.ctx.field(14):
            self.y = read_coord(s, self.ctx.deltaCoords, self.y)

        if self.ctx.field(15):
            cbData = Uint8.unpack(s)

            if cbData > 1:
                # Read glyph data.
                self.glyph = GlyphV2.parse(s)
                self.cacheIndex = self.glyph.cacheIndex
                s.read(2)  # Padding / Unicode representation
            else:
                # Only a cache index.
                assert cbData == 1
                self.glyph = None  # Glyph must be retrieved from cacheIndex
                self.cacheIndex = Uint8.unpack(s)

        return self

    def __str__(self):
        return f'<FastGlyph Cache={self.cacheId}:{self.cacheIndex} New={self.glyph != None}>'


class EllipseSc:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.left = 0
        self.top = 0
        self.right = 0
        self.bottom = 0
        self.rop2 = 0
        self.fillMode = 0
        self.color = 0

    def update(self, s: BytesIO):

        if self.ctx.field(1):
            self.left = read_coord(s, self.ctx.deltaCoords, self.left)
        if self.ctx.field(2):
            self.top = read_coord(s, self.ctx.deltaCoords, self.top)
        if self.ctx.field(3):
            self.right = read_coord(s, self.ctx.deltaCoords, self.right)
        if self.ctx.field(4):
            self.bottom = read_coord(s, self.ctx.deltaCoords, self.bottom)
        if self.ctx.field(5):
            self.rop2 = Uint8.unpack(s)
        if self.ctx.field(6):
            self.fillMode = Uint8.unpack(s)
        if self.ctx.field(7):
            self.color = read_rgb(s)

        return self

    def __str__(self):
        return '<EllipseSc>'


class EllipseCb:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx
        self.brush = Brush()

        self.left = 0
        self.top = 0
        self.right = 0
        self.bottom = 0
        self.rop2 = 0
        self.fillMode = 0
        self.bg = 0
        self.fg = 0

    def update(self, s: BytesIO):

        if self.ctx.field(1):
            self.left = read_coord(s, self.ctx.deltaCoords, self.left)
        if self.ctx.field(2):
            self.top = read_coord(s, self.ctx.deltaCoords, self.top)
        if self.ctx.field(3):
            self.right = read_coord(s, self.ctx.deltaCoords, self.right)
        if self.ctx.field(4):
            self.bottom = read_coord(s, self.ctx.deltaCoords, self.bottom)
        if self.ctx.field(5):
            self.rop2 = Uint8.unpack(s)
        if self.ctx.field(6):
            self.fillMode = Uint8.unpack(s)
        if self.ctx.field(7):
            self.bg = read_rgb(s)
        if self.ctx.field(8):
            self.fg = read_rgb(s)

        self.brush.update(s, self.ctx.fieldFlags >> 8)

        return self

    def __str__(self):
        return '<EllipseCb>'


class GlyphIndex:
    def __init__(self, ctx: PrimaryContext):
        self.ctx = ctx

        self.cacheId = 0
        self.flAccel = 0
        self.ulCharInc = 0
        self.fOpRedundant = 0
        self.bg = 0
        self.fg = 0
        self.bkLeft = 0
        self.bkTop = 0
        self.bkRight = 0
        self.bkBottom = 0
        self.opLeft = 0
        self.opTop = 0
        self.opRight = 0
        self.opBottom = 0

        self.brush = Brush()

        self.x = 0
        self.y = 0

        self.data = b''

    def update(self, s: BytesIO):

        if self.ctx.field(1):
            self.cacheId = Uint8.unpack(s)
        if self.ctx.field(2):
            self.flAccel = Uint8.unpack(s)
        if self.ctx.field(3):
            self.ulCharInc = Uint8.unpack(s)
        if self.ctx.field(4):
            self.fOpRedundant = Uint8.unpack(s)
        if self.ctx.field(5):
            self.bg = read_rgb(s)
        if self.ctx.field(6):
            self.fg = read_rgb(s)
        if self.ctx.field(7):
            self.bkLeft = Uint16LE.unpack(s)
        if self.ctx.field(8):
            self.bkTop = Uint16LE.unpack(s)
        if self.ctx.field(9):
            self.bkRight = Uint16LE.unpack(s)
        if self.ctx.field(10):
            self.bkBottom = Uint16LE.unpack(s)
        if self.ctx.field(11):
            self.opLeft = Uint16LE.unpack(s)
        if self.ctx.field(12):
            self.opTop = Uint16LE.unpack(s)
        if self.ctx.field(13):
            self.opRight = Uint16LE.unpack(s)
        if self.ctx.field(14):
            self.opBottom = Uint16LE.unpack(s)

        self.brush.update(s, self.ctx.fieldFlags >> 14)

        if self.ctx.field(20):
            self.x = Uint16LE.unpack(s)
        if self.ctx.field(21):
            self.y = Uint16LE.unpack(s)

        if self.ctx.field(22):
            cbData = Uint8.unpack(s)
            self.data = s.read(cbData)

        return self

    def __str__(self):
        return '<GlyphIndex>'



#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import functools
import logging
import platform
import time
from typing import Optional, Union

from PySide6.QtCore import QEvent, QObject, Qt
from PySide6.QtGui import QKeyEvent, QMouseEvent, QWheelEvent
from PySide6.QtWidgets import QWidget

from pyrdp.enum import MouseButton
from pyrdp.layer import PlayerLayer
from pyrdp.logging import LOGGER_NAMES
from pyrdp.pdu import PlayerBitmapPDU, PlayerForwardingStatePDU, PlayerKeyboardPDU, PlayerMouseButtonPDU, \
    PlayerMouseMovePDU, PlayerMouseWheelPDU, PlayerTextPDU
from pyrdp.player import keyboard
from pyrdp.player.keyboard import isRightControl
from pyrdp.player.QTimerSequencer import QTimerSequencer
from pyrdp.ui import QRemoteDesktop


class RDPMITMWidget(QRemoteDesktop):
    """
    RDP Widget that handles mouse and keyboard events and sends them to the MITM server.
    """

    KEY_SEQUENCE_DELAY = 0

    def __init__(self, width: int, height: int, layer: PlayerLayer, parent: Optional[QWidget] = None):
        super().__init__(width, height, parent = parent)
        self.layer = layer
        self.handleEvents = False
        self.log = logging.getLogger(LOGGER_NAMES.PLAYER)
        self.setFocusPolicy(Qt.FocusPolicy.ClickFocus)
        self.installEventFilter(self)


    def getTimetamp(self) -> int:
        return int(round(time.time() * 1000))


    def getMousePosition(self, event: Union[QMouseEvent, QWheelEvent]) -> (int, int):
        pos = event.position()
        return max(pos.x(), 0), max(pos.y(), 0)

    def mouseMoveEvent(self, event: QMouseEvent):
        if not self.handleEvents or not self.hasFocus():
            return

        x, y = self.getMousePosition(event)

        pdu = PlayerMouseMovePDU(self.layer.getCurrentTimeStamp(), x, y)
        self.layer.sendPDU(pdu)

    def mousePressEvent(self, event: QMouseEvent):
        if self.handleEvents:
            self.handleMouseButton(event, True)

    def mouseReleaseEvent(self, event: QMouseEvent):
        if self.handleEvents:
            self.handleMouseButton(event, False)

    def handleMouseButton(self, event: QMouseEvent, pressed: bool):
        x, y = self.getMousePosition(event)
        button = event.button()

        mapping = {
            Qt.MouseButton.LeftButton: MouseButton.LEFT_BUTTON,
            Qt.MouseButton.RightButton: MouseButton.RIGHT_BUTTON,
            Qt.MouseButton.MiddleButton: MouseButton.MIDDLE_BUTTON,
        }

        if button not in mapping:
            return

        pdu = PlayerMouseButtonPDU(self.layer.getCurrentTimeStamp(), x, y, mapping[button], pressed)
        self.layer.sendPDU(pdu)


    def wheelEvent(self, event: QWheelEvent):
        if not self.handleEvents:
            return

        x, y = self.getMousePosition(event)
        delta = event.delta()
        horizontal = event.orientation() == Qt.Orientation.Horizontal

        event.setAccepted(True)

        pdu = PlayerMouseWheelPDU(self.layer.getCurrentTimeStamp(), x, y, delta, horizontal)
        self.layer.sendPDU(pdu)


    # We need this to capture tab key events
    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        if self.handleEvents and event.type() == QEvent.KeyPress:
            self.keyPressEvent(event)
            return True

        return QObject.eventFilter(self, obj, event)


    def keyPressEvent(self, event: QKeyEvent):
        if not isRightControl(event):
            if self.handleEvents:
                self.handleKeyEvent(event, False)
        else:
            self.clearFocus()

    def keyReleaseEvent(self, event: QKeyEvent):
        if self.handleEvents:
            self.handleKeyEvent(event, True)

    def handleKeyEvent(self, event: QKeyEvent, released: bool):
        # After some testing, it seems like scan codes on Linux are 8 higher than their Windows version.
        if platform.system() == "Linux":
            offset = -8
        else:
            offset = 0

        scanCode = keyboard.findScanCodeForEvent(event) or event.nativeScanCode() + offset
        pdu = PlayerKeyboardPDU(self.layer.getCurrentTimeStamp(), scanCode, released, event.key() in keyboard.EXTENDED_KEYS)
        self.layer.sendPDU(pdu)


    def sendKeySequence(self, keys: [Qt.Key]):
        self.setFocus()

        pressPDUs = []
        releasePDUs = []

        for key in keys:
            scanCode = keyboard.SCANCODE_MAPPING[key]
            isExtended = key in keyboard.EXTENDED_KEYS

            pressPDU = PlayerKeyboardPDU(self.layer.getCurrentTimeStamp(), scanCode, False, isExtended)
            pressPDUs.append(pressPDU)

            releasePDU = PlayerKeyboardPDU(self.layer.getCurrentTimeStamp(), scanCode, True, isExtended)
            releasePDUs.append(releasePDU)

        def press() -> int:
            for pdu in pressPDUs:
                self.layer.sendPDU(pdu)

            return RDPMITMWidget.KEY_SEQUENCE_DELAY

        def release():
            for pdu in releasePDUs:
                self.layer.sendPDU(pdu)

        sequencer = QTimerSequencer([press, release])
        sequencer.run()


    def sendText(self, text: str):
        self.setFocus()

        functions = []

        def pressCharacter(character: str) -> int:
            pdu = PlayerTextPDU(self.layer.getCurrentTimeStamp(), character, False)
            print(c)
            self.layer.sendPDU(pdu)
            return RDPMITMWidget.KEY_SEQUENCE_DELAY

        def releaseCharacter(character: str):
            pdu = PlayerTextPDU(self.layer.getCurrentTimeStamp(), character, True)
            self.layer.sendPDU(pdu)

        for c in text:
            press = functools.partial(pressCharacter, c)
            release = functools.partial(releaseCharacter, c)
            functions.append(press)
            functions.append(release)

        sequencer = QTimerSequencer(functions)
        sequencer.run()


    def setControlState(self, controlled: bool):
        self.handleEvents = controlled
        self.setForwardingState(not controlled)

        if not controlled:
            self.sendCurrentScreen()

    def setForwardingState(self, shouldForward: bool):
        self.layer.sendPDU(PlayerForwardingStatePDU(self.layer.getCurrentTimeStamp(), shouldForward, shouldForward))

    def sendCurrentScreen(self):
        width = self._buffer.width()
        height = self._buffer.height()
        pdu = PlayerBitmapPDU(self.layer.getCurrentTimeStamp(), width, height, self._buffer.bits())
        self.layer.sendPDU(pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from PySide6.QtCore import QTimer

from pyrdp.core import Sequencer


class QTimerSequencer(Sequencer):
    """
    Sequencer that uses QTimer to wait between calls.
    """

    def wait(self, waitTime: int):
        QTimer.singleShot(waitTime, self.run)



#
# This file is part of the PyRDP project.
# Copyright (C) 2021, 2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
import enum
from typing import Tuple

from scapy.layers.inet import IP, TCP
from scapy.layers.l2 import Ether

from pyrdp.convert.JSONEventHandler import JSONEventHandler
from pyrdp.player import HAS_GUI

from pyrdp.convert.pyrdp_scapy import *

"""
Supported conversion handlers.

The class constructor signature must be `__init__(self, output_path: str, progress=None)`
"""
HANDLERS = {"replay": (None, "pyrdp"), "json": (JSONEventHandler, "json")}

if HAS_GUI:
    from pyrdp.convert.MP4EventHandler import MP4EventHandler
    HANDLERS["mp4"] = (MP4EventHandler, "mp4")
else:
    # Class stub for when MP4 support is not available.
    # It would be a good idea to refactor this so that Mp4EventHandler is
    # acquired through some factory object that checks for GUI support
    # once we add more conversion handlers.
    class MP4EventHandler():
        def __init__(self, _unused: str):
            pass


class TCPFlags(enum.IntEnum):
    FIN = 0x01
    SYN = 0x02
    RST = 0x04
    PSH = 0x08
    ACK = 0x10
    URG = 0x20
    ECE = 0x40
    CWR = 0x80


class InetAddress():
    def __init__(self, ip: str, port: int):
        self._ip = ip
        self._port = port

    @property
    def ip(self) -> str:
        return self._ip

    @property
    def port(self) -> int:
        return self._port

    def __eq__(self, other):
        if self.ip == other.ip and self.port == other.port:
            return True
        else:
            return False

    def __str__(self):
        return f"{self._ip}:{self._port}"


def extractInetAddressesFromPDUPacket(packet) -> Tuple[InetAddress, InetAddress]:
    """Returns the src and dst InetAddress (IP, port) from a PDU packet"""
    x = ExportedPDU(packet.load)
    return (InetAddress(x.src, x.sport), InetAddress(x.dst, x.dport))


def createHandler(format: str, outputFileBase: str, progress=None) -> Tuple[str, str]:
    """
    Gets the appropriate handler and returns the filename with extension.
    Returns None if the format is replay.
    TODO: Returning None if the format is replay is kind of janky. This could use a refactor to handle replays and other formats differently.
    """

    if format not in HANDLERS:
        print("[-] Unsupported conversion format.")
        sys.exit(1)

    HandlerClass, ext = HANDLERS[format]
    outputFileBase += f".{ext}"
    return HandlerClass(outputFileBase, progress=progress) if HandlerClass else None, outputFileBase


class ExportedPDU(Packet):
    """60 byte EXPORTED_PDU header."""
    # We could properly parse the EXPORTED_PDU struct, but we are mostly dealing with IP exported PDUs
    # so let's just wing it.
    name = "ExportedPDU"
    fields_desc = [ 
                    IntField("tag1Num", None),  # 4
                    StrFixedLenField("proto", None, length=4),  # 8
                    IntField("tag2Num", None),  # 12
                    IPField("src", None),  # 16
                    IntField("tag3Num", None),  # 20
                    IPField("dst", None),  # 24
                    IntField("tag4Num", None),  # 28
                    IntField("portType", None),  # 32
                    IntField("tag5Num", None),  # 36
                    IntField("sport", None),  # 40
                    IntField("tag6Num", None),  # 44
                    IntField("dport", None),  # 48
                    IntField("tag7Num", None),  # 52
                    IntField("frame", None),   # 56
                    IntField("endOfTags", None),  # 60
    ]


# noinspection PyUnresolvedReferences
def tcp_both(p) -> str:
    """Session extractor which merges both sides of a TCP channel."""

    if "TCP" in p:
        return str(
            sorted(["TCP", p[IP].src, p[TCP].sport, p[IP].dst, p[TCP].dport], key=str)
        )

    # Need to make sure this is OK when non-TCP, non-exported data is present.
    if Ether not in p:
        x = ExportedPDU(p.load)
        return str(
                sorted([x.proto.upper(), x.src, x.sport, x.dst, x.dport], key=str)
        )

    return "Unsupported"


# noinspection PyUnresolvedReferences
def findClientRandom(stream: PacketList, limit: int = 20) -> str:
    """Find the client random offset and value of a stream."""
    for n, p in enumerate(stream):
        if n >= limit:
            return ""  # Didn't find client hello.
        try:
            tls = p[TCP].payload
            hello = tls.msg[0]
            if isinstance(hello, TLSClientHello):
                return (pkcs_i2osp(hello.gmt_unix_time, 4) + hello.random_bytes).hex()
        except AttributeError as e:
            continue  # Not a TLS packet.

    return ""


def loadSecrets(filename: str) -> dict:
    secrets = {}
    with open(filename, "r") as f:
        for line in f:
            line = line.strip()
            if line == "" or not line.startswith("CLIENT"):
                continue

            parts = line.split(" ")
            if len(parts) != 3:
                continue

            [t, c, m] = parts

            # Parse the secret accordingly.
            if t == "CLIENT_RANDOM":
                secrets[c] = {"client": bytes.fromhex(c), "master": bytes.fromhex(m)}
    return secrets

def canExtractSessionInfo(session: PacketList) -> bool:
    packet = session[0]
    # TODO: Eventually we should be able to wrap the session as an ExportedSession
    # and check for the presence of exported.
    return IP in packet or Ether not in packet

def getSessionInfo(session: PacketList) -> Tuple[InetAddress, InetAddress, float, bool]:
    """Attempt to retrieve an (src, dst, ts, isPlaintext) tuple for a data stream."""
    packet = session[0]

    # FIXME: This relies on the fact that decrypted traces are using EXPORTED_PDU and
    #        thus have no `Ether` layer, but it is technically possible to have a true
    #        plaintext capture with very old implementations of RDP.
    if TCP in packet:
        # Assume an encrypted stream...
        return (InetAddress(packet[IP].src, packet[IP][TCP].sport),
                InetAddress(packet[IP].dst, packet[IP][TCP].dport),
                packet.time, False)
    elif Ether not in packet:
        # No Ethernet layer, so assume exported PDUs.
        src, dst = extractInetAddressesFromPDUPacket(packet)
        return (src, dst, packet.time, True)

    raise Exception("Invalid stream type. Must be TCP/TLS or EXPORTED PDU.")


#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from abc import ABCMeta, abstractmethod
from typing import List, Union

from pyrdp.core import EventEngine, ObservedBy, Observer, Subject
from pyrdp.exceptions import UnknownPDUTypeError, ParsingError
from pyrdp.parser import Parser
from pyrdp.pdu import PDU


class LayerObserver(Observer):
    """
    Layer observer class, notified whenever a layer receives a PDU.
    """
    def __init__(self, **kwargs):
        Observer.__init__(self, **kwargs)

    def onPDUReceived(self, pdu):
        """
        Method called when a PDU is received
        """
        pass


class LayerRoutedObserver(LayerObserver):
    """
    Layer observer that routes PDUs to methods by checking the PDU's header.
    """
    def __init__(self, handlers: dict, **kwargs):
        """
        :param handlers: a dictionary of headers to method names
        """
        LayerObserver.__init__(self, **kwargs)
        self.handlers = {}

        for (header, name) in handlers.items():
            self.handlers[header] = getattr(self, name)

    def onPDUReceived(self, pdu):
        if pdu.header not in self.handlers:
            self.onUnknownHeader(pdu)
        else:
            self.handlers[pdu.header](pdu)

    def onUnknownHeader(self, pdu):
        """
        Method called when a PDU with an unknown header is received
        """
        pass


class LayerStrictRoutedObserver(LayerRoutedObserver):
    """
    Layer observer that throws an exception when an unknown header is received.
    """
    def __init__(self, handlers: dict, **kwargs):
        """
        :param handlers: a dictionary of headers to method names
        """
        LayerRoutedObserver.__init__(self, handlers, **kwargs)
    
    def onUnknownHeader(self, pdu):
        """
        Method called when a PDU with an unknown header is received
        """
        raise UnknownPDUTypeError("Unknown PDU header received: 0x%2lx" % pdu.header, type(pdu))


class ByteSender(metaclass = ABCMeta):
    """
    Basic interface that defines the sendBytes method.
    """

    @abstractmethod
    def sendBytes(self, data: bytes):
        pass


@ObservedBy(LayerObserver)
class BaseLayer(Subject):
    """
    Transforms bytes into PDUs. This is the most basic interface for layers. This is mostly for layers that
    don't use parsers to handle the data they receive (see SegmentationLayer).
    Parsed PDUs are processed by the event engine.
    Observers can be attached to be notified of incoming PDUs.
    ObservedBy: LayerObserver
    """

    def __init__(self):
        super().__init__()
        self.eventEngine = EventEngine()
        self.previous: ByteSender = None

    def setPrevious(self, previous: ByteSender):
        self.previous = previous

    def pduReceived(self, pdu: PDU):
        """
        Call when a PDU is received to have it processed (notify the observer and pass it through the event engine).
        :param pdu: the PDU.
        """
        self.eventEngine.processObject(pdu)

        if self.observer is not None:
            self.observer.onPDUReceived(pdu)

    def recv(self, data: bytes):
        """
        Parse received bytes to a PDU and process it.
        :param data: bytes received.
        """
        raise NotImplementedError("recv must be overridden")

    async def waitPDU(self, *args, **kwargs) -> PDU:
        """
        Wait for a PDU matching certain criteria (see EventEngine.wait).
        """
        return await self.eventEngine.wait(*args, **kwargs)


class Layer(BaseLayer):
    """
    A layer transforms bytes into PDUs by using a given parser.
    Parsed PDUs are processed by the event engine.
    Observers can be attached to be notified of incoming PDUs.
    ObservedBy: LayerObserver
    """
    def __init__(self, mainParser: Parser):
        super().__init__()
        self.mainParser = mainParser

    def recv(self, data: bytes):
        """
        Parse received bytes to a PDU and process it.
        :param data: bytes received.
        """
        pdu = self.mainParser.parse(data)

        try:
            self.pduReceived(pdu)
        except ParsingError as e:
            e.addLayer(self.mainParser, data)
            raise

    def sendPDU(self, pdu: PDU):
        """
        Send a PDU to the previous layer.
        :param pdu: the PDU.
        """
        data = self.mainParser.write(pdu)
        self.previous.sendBytes(data)


class LayerChainItem(ByteSender, metaclass=ABCMeta):
    def __init__(self):
        super().__init__()
        self.next: BaseLayer = None

    @staticmethod
    def chain(first: 'LayerChainItem', second: Union['BaseLayer', 'LayerChainItem'], *layers: List[Union['BaseLayer', 'LayerChainItem']]):
        """
        Chain a series of layers together by calling setNext iteratively.
        :param first: first layer in the chain.
        :param second: second layer in the chain.
        :param layers: additional layers in the chain.
        """
        first.setNext(second)

        current = second
        for nextLayer in layers:
            current.setNext(nextLayer)
            current = nextLayer

    def setNext(self, layer: 'BaseLayer'):
        """
        Set the next layer in the protocol hierarchy (ex: IP's next layer would be TCP/UDP).
        :param layer: The next layer.
        """
        self.next = layer
        layer.setPrevious(self)


class IntermediateLayer(Layer, LayerChainItem):
    """
    Layer that usually forwards some or all PDUs to another layer.
    This is mostly for layers that are not "endpoint" layers (i.e: they only process part of the data).
    TPKTLayer is an example of an IntermediateLayer.
    SlowPathLayer is an example of an "endpoint" layer that doesn't forward anything.
    """

    def __init__(self, mainParser: Parser):
        super().__init__(mainParser)

    def pduReceived(self, pdu: PDU):
        Layer.pduReceived(self, pdu)

        if self.next is not None and self.shouldForward(pdu):
            self.next.recv(pdu.payload)

    def sendBytes(self, data: bytes):
        self.previous.sendBytes(data)

    def shouldForward(self, pdu: PDU) -> bool:
        """
        Determine if a PDU should be forwarded to the next layer or not.
        :param pdu: the PDU.
        :return: True if the PDU should be forwarded.
        """
        raise NotImplementedError("shouldForward must be overridden")


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer.layer import Layer
from pyrdp.parser import ClipboardParser


class ClipboardLayer(Layer):
    """
    Layer to manage reception of the clipboard virtual channel data.
    https://msdn.microsoft.com/en-us/library/cc241066.aspx
    """

    def __init__(self, parser = ClipboardParser()):
        super().__init__(parser)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from logging import LoggerAdapter

from pyrdp.layer import FastPathLayer
from pyrdp.logging.StatCounter import StatCounter, STAT
from pyrdp.mitm.state import RDPMITMState
from pyrdp.pdu import FastPathPDU, FastPathScanCodeEvent, FastPathMouseEvent
from pyrdp.enum import scancode
from pyrdp.mitm.BasePathMITM import BasePathMITM

class FastPathMITM(BasePathMITM):
    """
    MITM component for the fast-path layer.
    """

    def __init__(self, client: FastPathLayer, server: FastPathLayer, state: RDPMITMState, statCounter: StatCounter, log: LoggerAdapter):
        """
        :param client: fast-path layer for the client side
        :param server: fast-path layer for the server side
        :param state: the MITM state.
        """

        super().__init__(state, client, server, statCounter, log)

        self.client.createObserver(
            onPDUReceived=self.onClientPDUReceived,
        )

        self.server.createObserver(
            onPDUReceived=self.onServerPDUReceived,
        )

    def onClientPDUReceived(self, pdu: FastPathPDU):
        self.statCounter.increment(STAT.IO_INPUT_FASTPATH)
        if self.state.forwardInput:
            self.server.sendPDU(pdu)

        if not self.state.loggedIn:
            for event in pdu.events:
                if isinstance(event, FastPathScanCodeEvent):
                    self.onScanCode(event.scanCode, event.isReleased, event.rawHeaderByte & scancode.KBDFLAGS_EXTENDED != 0)
                elif isinstance(event, FastPathMouseEvent):
                    self.onMouse(event.mouseX, event.mouseY, event.pointerFlags)

    def onServerPDUReceived(self, pdu: FastPathPDU):
        self.statCounter.increment(STAT.IO_OUTPUT_FASTPATH)
        if self.state.forwardOutput:
            self.client.sendPDU(pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.layer import AsyncIOTCPLayer, LayerChainItem, PlayerLayer, TwistedTCPLayer


class TwistedPlayerLayerSet:
    def __init__(self):
        self.tcp = TwistedTCPLayer()
        self.player = PlayerLayer()
        LayerChainItem.chain(self.tcp, self.player)


class AsyncIOPlayerLayerSet:
    def __init__(self):
        self.tcp = AsyncIOTCPLayer()
        self.player = PlayerLayer()
        LayerChainItem.chain(self.tcp, self.player)


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import List

from pyrdp.enum import DeviceType, PlayerPDUType
from pyrdp.enum.player import MouseButton
from pyrdp.pdu.pdu import PDU


class PlayerPDU(PDU):
    """
    PDU to encapsulate different types (ex: input, output, creds) for (re)play purposes.
    Also contains a timestamp.
    """

    def __init__(self, header: PlayerPDUType, timestamp: int, payload: bytes):
        self.header = header  # Uint16LE
        self.timestamp = timestamp  # Uint64LE
        PDU.__init__(self, payload)


class PlayerConnectionClosePDU(PlayerPDU):
    def __init__(self, timestamp: int):
        super().__init__(PlayerPDUType.CONNECTION_CLOSE, timestamp, b"")


class PlayerMouseMovePDU(PlayerPDU):
    """
    PDU definition for mouse move events coming from the player.
    """

    def __init__(self, timestamp: int, x: int, y: int):
        super().__init__(PlayerPDUType.MOUSE_MOVE, timestamp, b"")
        self.x = x
        self.y = y


class PlayerMouseButtonPDU(PlayerPDU):
    """
    PDU definition for mouse button events coming from the player.
    """

    def __init__(self, timestamp: int, x: int, y: int, button: MouseButton, pressed: bool):
        super().__init__(PlayerPDUType.MOUSE_BUTTON, timestamp, b"")
        self.x = x
        self.y = y
        self.button = button
        self.pressed = pressed


class PlayerMouseWheelPDU(PlayerPDU):
    """
    PDU definition for mouse wheel events coming from the player.
    """

    def __init__(self, timestamp: int, x: int, y: int, delta: int, horizontal: bool):
        super().__init__(PlayerPDUType.MOUSE_WHEEL, timestamp, b"")
        self.x = x
        self.y = y
        self.delta = delta
        self.horizontal = horizontal


class PlayerKeyboardPDU(PlayerPDU):
    """
    PDU definition for keyboard events coming from the player.
    """

    def __init__(self, timestamp: int, code: int, released: bool, extended: bool):
        super().__init__(PlayerPDUType.KEYBOARD, timestamp, b"")
        self.code = code
        self.released = released
        self.extended = extended


class PlayerTextPDU(PlayerPDU):
    """
    PDU definition for text events coming from the player.
    """

    def __init__(self, timestamp: int, character: str, released: bool):
        super().__init__(PlayerPDUType.TEXT, timestamp, b"")
        self.character = character
        self.released = released


class PlayerForwardingStatePDU(PlayerPDU):
    """
    PDU definition for changing the state of I/O forwarding.
    """

    def __init__(self, timestamp: int, forwardInput: bool, forwardOutput: bool):
        super().__init__(PlayerPDUType.FORWARDING_STATE, timestamp, b"")
        self.forwardInput = forwardInput
        self.forwardOutput = forwardOutput


class Color:
    def __init__(self, r: int, g: int, b: int, a: int):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

class PlayerBitmapPDU(PlayerPDU):
    """
    PDU definition for bitmap events.
    """

    def __init__(self, timestamp: int, width: int, height: int, pixels: bytes):
        """
        :param timestamp: timestamp.
        :param width: bitmap width.
        :param height: bitmap height.
        :param pixels: Array of colors organized in a left to right, top to bottom fashion: [(x0, y0), (x1, y0), ..., (x0, y1), (x1, y1)].
        """

        super().__init__(PlayerPDUType.BITMAP, timestamp, b"")
        self.width = width
        self.height = height
        self.pixels = pixels

    def __repr__(self):
        properties = dict(self.__dict__)
        properties["pixels"] = f"[Color * {len(self.pixels)}]"
        representation = self.__class__.__name__ + str(properties)
        return representation


class PlayerDeviceMappingPDU(PlayerPDU):
    def __init__(self, timestamp: int, deviceID: int, deviceType: DeviceType, name: str):
        super().__init__(PlayerPDUType.DEVICE_MAPPING, timestamp, b"")
        self.deviceID = deviceID
        self.deviceType = deviceType
        self.name = name


class PlayerDirectoryListingRequestPDU(PlayerPDU):
    def __init__(self, timestamp: int, deviceID: int, path: str):
        """
        :param timestamp: time stamp for this PDU.
        :param deviceID: ID of the device used.
        :param path: path of the directory to list. The path should be a Unix-style path.
        """

        super().__init__(PlayerPDUType.DIRECTORY_LISTING_REQUEST, timestamp, b"")
        self.deviceID = deviceID
        self.path = path


class PlayerFileDescription(PDU):
    def __init__(self, filePath: str, isDirectory: bool):
        """
        :param filePath: Unix-style path of the file.
        :param isDirectory: True if the file is a directory.
        """

        super().__init__()
        self.path = filePath
        self.isDirectory = isDirectory


class PlayerDirectoryListingResponsePDU(PlayerPDU):
    def __init__(self, timestamp: int, deviceID: int, fileDescriptions: List[PlayerFileDescription]):
        """
        :param timestamp: time stamp for this PDU.
        :param deviceID: ID of the device used.
        :param fileDescriptions: list of file descriptions.
        """

        super().__init__(PlayerPDUType.DIRECTORY_LISTING_RESPONSE, timestamp, b"")
        self.deviceID = deviceID
        self.fileDescriptions = fileDescriptions


class PlayerFileDownloadRequestPDU(PlayerPDU):
    def __init__(self, timestamp: int, deviceID: int, path: str):
        """
        :param timestamp: time stamp for this PDU.
        :param deviceID: ID of the device used.
        :param path: path of the directory to list. The path should be a Unix-style path.
        """

        super().__init__(PlayerPDUType.FILE_DOWNLOAD_REQUEST, timestamp, b"")
        self.deviceID = deviceID
        self.path = path

class PlayerFileDownloadResponsePDU(PlayerPDU):
    def __init__(self, timestamp: int, deviceID: int, path: str, offset: int, payload: bytes):
        """
        :param timestamp: time stamp for this PDU.
        :param deviceID: ID of the device used.
        :param path: path of the directory to list. The path should be a Unix-style path.
        :param offset: offset at which the data starts in the file.
        :param payload: file data that was read.
        """

        super().__init__(PlayerPDUType.FILE_DOWNLOAD_RESPONSE, timestamp, payload)
        self.deviceID = deviceID
        self.path = path
        self.offset = offset

class PlayerFileDownloadCompletePDU(PlayerPDU):
    def __init__(self, timestamp: int, deviceID: int, path: str, error: int):
        """
        :param timestamp: time stamp for this PDU.
        :param deviceID: ID of the device used.
        :param path: path of the directory to list. The path should be a Unix-style path.
        :param error: error that resulted in completion (0 if nothing went wrong).
        """

        super().__init__(PlayerPDUType.FILE_DOWNLOAD_COMPLETE, timestamp, b"")
        self.deviceID = deviceID
        self.path = path
        self.error = error


#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import time

from pyrdp.core import ObservedBy
from pyrdp.enum import PlayerPDUType
from pyrdp.layer import BufferedLayer, LayerRoutedObserver
from pyrdp.parser import PlayerParser
from pyrdp.pdu import PlayerPDU

class PlayerLayer(BufferedLayer):
    """
    Layer to manage the encapsulation of Player metadata such as event timestamp and
    event type/origin (input, output).
    """

    def __init__(self, parser: PlayerParser = PlayerParser()):
        super().__init__(parser)

    def sendMessage(self, data: bytes, messageType: PlayerPDUType, timeStamp: int):
        pdu = PlayerPDU(messageType, timeStamp, data)
        self.sendPDU(pdu)

    @staticmethod
    def timeStampFunction() -> int:
        return int(time.time() * 1000)

    def getCurrentTimeStamp(self) -> int:
        return PlayerLayer.timeStampFunction()


# This file is part of the PyRDP project.
#
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

class ImageHandler:
    def notifyImage(self, x: int, y: int, img: 'QImage', width: int, height: int):
        raise NotImplementedError("ImageHandler.notifyImage is not implemented.")

    def resize(self, width: int, height: int):
        raise NotImplementedError("ImageHandler.resize is not implemented.")

    def update(self) -> int:
        raise NotImplementedError("ImageHandler.update is not implemented")

    @property
    def screen(self) -> 'QImage':
        raise NotImplementedError("ImageHandler.screen is not implemented")



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Uint16LE, Uint32LE, Uint8
from pyrdp.enum import FIPSVersion, SecurityFlags
from pyrdp.parser.parser import Parser
from pyrdp.pdu import SecurityExchangePDU, SecurityPDU
from pyrdp.security import RC4Crypter, RC4CrypterProxy


class BasicSecurityParser(Parser):
    """
    Base class for all security parsers.
    This class only reads a small header before the payload.
    Writing is split between 3 methods for reusability.
    """

    def doParse(self, data):
        """
        Decode a security PDU from bytes.
        :type data: bytes
        :return: RDPSecurityPDU
        """
        stream = BytesIO(data)
        header = Uint32LE.unpack(stream)

        if header & SecurityFlags.SEC_EXCHANGE_PKT != 0:
            return self.parseSecurityExchange(stream, header)

        payload = stream.read()
        return SecurityPDU(header, payload)

    def parseSecurityExchange(self, stream, header):
        """
        Decode a security exchange PDU.
        :type stream: BytesIO
        :type header: int
        :return: RDPSecurityExchangePDU
        """
        length = Uint32LE.unpack(stream)
        clientRandom = stream.read(length)
        return SecurityExchangePDU(header, clientRandom)

    def write(self, pdu):
        """
        Encode a security PDU to bytes.
        :type pdu: SecurityPDU
        :return: str
        """
        stream = BytesIO()
        self.writeHeader(stream, pdu)
        self.writeBody(stream, pdu)
        self.writePayload(stream, pdu)
        return stream.getvalue()

    def writeSecurityExchange(self, pdu):
        """
        Encode a RDPSecurityExchangePDU to bytes.
        :type pdu: SecurityExchangePDU
        :return: str
        """
        stream = BytesIO()
        Uint32LE.pack(SecurityFlags.SEC_EXCHANGE_PKT | SecurityFlags.SEC_LICENSE_ENCRYPT_SC, stream)
        Uint32LE.pack(len(pdu.clientRandom), stream)
        stream.write(pdu.clientRandom)
        return stream.getvalue()

    def writeHeader(self, stream, pdu):
        """
        Write the PDU header.
        :type stream: BytesIO
        :type pdu: SecurityPDU
        """
        Uint32LE.pack(pdu.header, stream)

    def writeBody(self, stream, pdu):
        """
        Write the PDU body.
        :type stream: BytesIO
        :type pdu: SecurityPDU
        """
        pass

    def writePayload(self, stream, pdu):
        """
        Write the PDU payload.
        :type stream: BytesIO
        :type pdu: SecurityPDU
        """
        stream.write(pdu.payload)



class SignedSecurityParser(BasicSecurityParser):
    """
    Parser to use when standard RDP security is used.
    This class handles RC4 decryption and encryption and increments the operation count automatically.
    """

    def __init__(self, crypter):
        """
        :type crypter: RC4Crypter | RC4CrypterProxy
        """
        BasicSecurityParser.__init__(self)
        self.crypter = crypter

    def parse(self, data):
        stream = BytesIO(data)
        header = Uint32LE.unpack(stream)

        if header & SecurityFlags.SEC_EXCHANGE_PKT != 0:
            return self.parseSecurityExchange(stream, header)

        signature = stream.read(8)
        payload = stream.read()

        if header & SecurityFlags.SEC_ENCRYPT != 0:
            payload = self.crypter.decrypt(payload)
            self.crypter.addDecryption()

        return SecurityPDU(header, payload)


    def writeHeader(self, stream, pdu):
        # Make sure the header contains the flags for encryption and salted signatures.
        header = pdu.header | SecurityFlags.SEC_ENCRYPT | SecurityFlags.SEC_SECURE_CHECKSUM
        Uint32LE.pack(header, stream)

    def writeBody(self, stream, pdu):
        # Write the signature before writing the payload.
        signature = self.crypter.sign(pdu.payload, True)
        stream.write(signature)

    def writePayload(self, stream, pdu):
        payload = self.crypter.encrypt(pdu.payload)
        self.crypter.addEncryption()
        stream.write(payload)



class FIPSSecurityParser(SignedSecurityParser):
    """
    Parser to use when FIPS security is used.
    Note that FIPS cryptography is not implemented yet.
    """

    def __init__(self, crypter):
        """
        :type crypter: RC4Crypter | RC4CrypterProxy
        """
        SignedSecurityParser.__init__(self, crypter)

    def parse(self, data):
        stream = BytesIO(data)
        header = Uint32LE.unpack(stream)

        if header & SecurityFlags.SEC_EXCHANGE_PKT != 0:
            return self.parseSecurityExchange(stream, header)

        length = Uint16LE.unpack(stream)
        version = Uint8.unpack(stream)
        padLength = Uint8.unpack(stream)
        signature = stream.read(8)
        payload = stream.read()

        if header & SecurityFlags.SEC_ENCRYPT != 0:
            payload = self.crypter.decrypt(payload)
            self.crypter.addDecryption()

        return SecurityPDU(header, payload)

    def writeBody(self, stream, pdu):
        Uint16LE.pack(0x10, stream)
        Uint8.pack(FIPSVersion.TSFIPS_VERSION1, stream)
        Uint8.pack(self.crypter.getPadLength(pdu.payload), stream)
        SignedSecurityParser.writeBody(self, stream, pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2020-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
File that contains methods related to the MITM command line.
To be consumed either via bin/pyrdp-mitm.py or via twistd plugin.
"""
import argparse
import logging
import logging.handlers
import os
import sys
from base64 import b64encode
from pathlib import Path
from typing import Tuple

import OpenSSL

from pyrdp.core import settings
from pyrdp.core.ssl import ServerTLSContext
from pyrdp.logging import configure as configureLoggers, LOGGER_NAMES
from pyrdp.mitm.config import DEFAULTS, MITMConfig
from pyrdp.enum import NegotiationProtocols


def parseTarget(target: str) -> Tuple[str, int]:
    """
    Parse a target host:port and return components. Port is optional.
    """
    if ":" in target:
        targetHost = target[: target.index(":")]
        targetPort = int(target[target.index(":") + 1:])
    else:
        targetHost = target
        targetPort = 3389
    return targetHost, targetPort


def validateKeyAndCertificate(private_key: str, certificate: str) -> Tuple[str, str]:
    if (private_key is None) != (certificate is None):
        sys.stderr.write("You must provide both the private key and the certificate\n")
        sys.exit(1)

    if private_key is None:
        # Certificates will be generated automatically.
        return None, None
    else:
        key, cert = private_key, certificate

    try:
        # Check if OpenSSL accepts the private key and certificate.
        ServerTLSContext(key, cert)
    except OpenSSL.SSL.Error as error:
        from pyrdp.logging import log
        log.error(
            "An error occurred when creating the server TLS context. " +
            "There may be a problem with your private key or certificate (e.g: signature algorithm too weak). " +
            "Here is the exception: %(error)s",
            {"error": error}
        )
        sys.exit(1)

    return key, cert


def showConfiguration(config: MITMConfig):
    logging.getLogger(LOGGER_NAMES.MITM).info("Target: %(target)s:%(port)d", {
        "target": config.targetHost, "port": config.targetPort})
    logging.getLogger(LOGGER_NAMES.MITM).info("Output directory: %(outputDirectory)s",
                                              {"outputDirectory": config.outDir.absolute()})


def buildArgParser():
    parser = argparse.ArgumentParser()
    parser.add_argument("target", help="IP:port of the target RDP machine (ex: 192.168.1.10:3390)", nargs='?', default=None)
    parser.add_argument("-a", "--address", help="Address to listen on (default: 0.0.0.0)", default="0.0.0.0")
    parser.add_argument("-l", "--listen", help="Port number to listen on (default: 3389)", default=3389)
    parser.add_argument("-o", "--output", help="Output folder", default="pyrdp_output")
    parser.add_argument("-i", "--destination-ip",
                        help="Destination IP address of the PyRDP player.If not specified, RDP events are"
                        " not sent over the network.")
    parser.add_argument("-d", "--destination-port",
                        help="Listening port of the PyRDP player (default: 3000).", default=3000)
    parser.add_argument("-k", "--private-key", help="Specify path to private key (for SSL)")
    parser.add_argument("-c", "--certificate", help="Specify path to certificate (for SSL).")
    parser.add_argument("-u", "--username", help="Username that will replace the client's username", default=None)
    parser.add_argument("-p", "--password", help="Password that will replace the client's password", default=None)
    parser.add_argument("-L", "--log-level", help="Console logging level. Logs saved to file are always verbose.",
                        default="INFO", choices=["INFO", "DEBUG", "WARNING", "ERROR", "CRITICAL"])
    parser.add_argument("-F", "--log-filter",
                        help="Only show logs from this logger name (accepts '*' wildcards)", default="")
    parser.add_argument("--auth", help="Specify allowed authentication mechanisms (Comma-separated, choose from: tls, ssp)", default="tls")
    parser.add_argument(
        "-s", "--sensor-id", help="Sensor ID (to differentiate multiple instances of the MITM"
        " where logs are aggregated at one place)")
    parser.add_argument("--payload", help="Command to run automatically upon connection", default=None)
    parser.add_argument("--payload-powershell",
                        help="PowerShell command to run automatically upon connection", default=None)
    parser.add_argument("--payload-powershell-file",
                        help="PowerShell script to run automatically upon connection (as -EncodedCommand)",
                        default=None)
    parser.add_argument(
        "--payload-delay", help="Time to wait after a new connection before sending the payload, in milliseconds",
        default=None)
    parser.add_argument(
        "--payload-duration", help="Amount of time for which input / output should be dropped, in milliseconds."
        " This can be used to hide the payload screen.", default=None)
    parser.add_argument("--disable-active-clipboard",
                        help="Disables the active clipboard stealing to request clipboard content upon connection.",
                        action="store_true")
    parser.add_argument("--crawl", help="Enable automatic shared drive scraping", action="store_true")
    parser.add_argument("--crawler-match-file",
                        help="File to be used by the crawler to chose what to download when scraping the client shared"
                        " drives.", default=None)
    parser.add_argument("--crawler-ignore-file",
                        help="File to be used by the crawler to chose what folders to avoid when scraping the client"
                        " shared drives.", default=None)
    parser.add_argument("--no-replay", help="Disable replay recording", action="store_true")
    parser.add_argument("--no-downgrade", help="Disables downgrading of unsupported extensions. This makes PyRDP harder"
                        " to fingerprint but might impact the player's ability to replay captured traffic.",
                        action="store_true")
    parser.add_argument(
        "--no-files", help="Do not extract files transferred between the client and server.", action="store_true")
    parser.add_argument(
        "--transparent", help="Spoof source IP for connections to the server (See README)", action="store_true")
    parser.add_argument("--no-gdi", help="Disable accelerated graphics pipeline (MS-RDPEGDI) extension",
                        action="store_true")
    parser.add_argument("--nla-redirection-host", help="Redirection target ip if NLA is enforced", default=None)
    parser.add_argument("--nla-redirection-port", help="Redirection target port if NLA is enforced", type=int, default=None)
    parser.add_argument("--ssp-challenge", help="Set challenge for SSP authentictation (e.g. 1122334455667788). Incompatible with --auth ssp.", type=str, default=None)

    return parser


def configure(cmdline=None) -> MITMConfig:
    parser = buildArgParser()

    if cmdline:
        args = parser.parse_args(cmdline)
    else:
        args = parser.parse_args()

    # Load configuration file.
    cfg = settings.load(settings.CONFIG_DIR + '/mitm.ini', DEFAULTS)

    # Override some of the switches based on command line arguments.
    if args.output:
        cfg.set('vars', 'output_dir', args.output)
    if args.log_filter:
        cfg.set('logs', 'filter', args.log_filter)
    if args.log_level:
        cfg.set('vars', 'level', args.log_level)
    if args.sensor_id:
        cfg.set('vars', 'sensor_id', args.sensor_id)

    outDir = Path(cfg.get('vars', 'output_dir'))
    outDir.mkdir(exist_ok=True)

    configureLoggers(cfg)
    logger = logging.getLogger(LOGGER_NAMES.PYRDP)

    if args.target is None and not args.transparent:
        parser.print_usage()
        sys.stderr.write('error: A relay target is required unless running in transparent proxy mode.\n')
        sys.exit(1)

    if (args.nla_redirection_host is None) != (args.nla_redirection_port is None):
        sys.stderr.write('Error: please provide both --nla-redirection-host and --nla-redirection-port\n')
        sys.exit(1)

    if args.ssp_challenge is not None and "ssp" in args.auth:
        sys.stderr.write('Error: Using a fixed challenge does not work with --auth ssp which is meant to specify an NLA bypass. '
                         'Without --auth ssp, an authentication downgrade attack will be attempted and if it is not possible, '
                         'the hash will be stolen with the fixed challenge.\n')
        sys.exit(1)

    if args.ssp_challenge is not None and len(args.ssp_challenge) != 16:
        sys.stderr.write('Error: please provide a challenge with 16 characters (64 bits)\n')
        sys.exit(1)

    if args.target:
        targetHost, targetPort = parseTarget(args.target)
    else:
        targetHost = None
        targetPort = 3389  # FIXME: Allow to set transparent port as well.

    key, certificate = validateKeyAndCertificate(args.private_key, args.certificate)

    config = MITMConfig()
    config.targetHost = targetHost
    config.targetPort = targetPort
    config.privateKeyFileName = key
    config.listenAddress = args.address
    config.listenPort = int(args.listen)
    config.certificateFileName = certificate
    config.attackerHost = args.destination_ip
    config.attackerPort = int(args.destination_port)
    config.replacementUsername = args.username
    config.replacementPassword = args.password
    config.outDir = outDir
    config.enableCrawler = args.crawl
    config.crawlerMatchFileName = args.crawler_match_file
    config.crawlerIgnoreFileName = args.crawler_ignore_file
    config.recordReplays = not args.no_replay
    config.downgrade = not args.no_downgrade
    config.transparent = args.transparent
    config.extractFiles = not args.no_files
    config.disableActiveClipboardStealing = args.disable_active_clipboard
    config.useGdi = not args.no_gdi
    config.redirectionHost = args.nla_redirection_host
    config.redirectionPort = args.nla_redirection_port
    config.sspChallenge = args.ssp_challenge

    payload = None
    powershell = None

    npayloads = int(args.payload is not None) + \
        int(args.payload_powershell is not None) + \
        int(args.payload_powershell_file is not None)

    if npayloads > 1:
        logger.error("Only one of --payload, --payload-powershell and --payload-powershell-file may be supplied.")
        sys.exit(1)

    if args.payload is not None:
        payload = args.payload
        logger.info("Using payload: %(payload)s", {"payload": args.payload})
    elif args.payload_powershell is not None:
        powershell = args.payload_powershell
        logger.info("Using powershell payload: %(payload)s", {"payload": args.payload_powershell})
    elif args.payload_powershell_file is not None:
        if not os.path.exists(args.payload_powershell_file):
            logger.error("Powershell file %(path)s does not exist.", {"path": args.payload_powershell_file})
            sys.exit(1)

        try:
            with open(args.payload_powershell_file, "r") as f:
                powershell = f.read()
        except IOError as e:
            logger.error("Error when trying to read powershell file: %(error)s", {"error": e})
            sys.exit(1)

        logger.info("Using payload from powershell file: %(path)s", {"path": args.payload_powershell_file})

    if powershell is not None:
        payload = "powershell -EncodedCommand " + b64encode(powershell.encode("utf-16le")).decode()

    if payload is not None:
        if args.payload_delay is None:
            logger.error("--payload-delay must be provided if a payload is provided.")
            sys.exit(1)

        if args.payload_duration is None:
            logger.error("--payload-duration must be provided if a payload is provided.")
            sys.exit(1)

        try:
            config.payloadDelay = int(args.payload_delay)
        except ValueError:
            logger.error("Invalid payload delay. Payload delay must be an integral number of milliseconds.")
            sys.exit(1)

        if config.payloadDelay < 0:
            logger.error("Payload delay must not be negative.")
            sys.exit(1)

        if config.payloadDelay < 1000:
            logger.warning(
                "You have provided a payload delay of less than 1 second."
                " We recommend you use a slightly longer delay to make sure it runs properly.")

        try:
            config.payloadDuration = int(args.payload_duration)
        except ValueError:
            logger.error("Invalid payload duration. Payload duration must be an integral number of milliseconds.")
            sys.exit(1)

        if config.payloadDuration < 0:
            logger.error("Payload duration must not be negative.")
            sys.exit(1)

        config.payload = payload
    elif args.payload_delay is not None:
        logger.error("--payload-delay was provided but no payload was set.")
        sys.exit(1)

    # Configure allowed authentication protocols.
    for auth in args.auth.split(','):
        auth = auth.strip()
        if auth == "tls":
            config.authMethods |= NegotiationProtocols.SSL
        elif auth == "ssp":
            # CredSSP implies TLS.
            config.authMethods |= (NegotiationProtocols.SSL | NegotiationProtocols.CRED_SSP)

    showConfiguration(config)
    return config



#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
Constants, state and parsing primitives for Alternate Secondary Drawing Orders.
"""
from io import BytesIO

from pyrdp.core import Uint16LE, Uint8, Uint32LE
from .common import read_color

STREAM_BITMAP_END = 0x01
STREAM_BITMAP_COMPRESSED = 0x02
STREAM_BITMAP_V2 = 0x04


class CreateOffscreenBitmap:
    @staticmethod
    def parse(s: BytesIO) -> 'CreateOffscreenBitmap':
        self = CreateOffscreenBitmap()

        self.flags = Uint16LE.unpack(s)
        self.id = self.flags & 0x7FFF
        self.cx = Uint16LE.unpack(s)
        self.cy = Uint16LE.unpack(s)

        if self.flags & 0x8000 != 0:
            cIndices = Uint16LE.unpack(s)
            self.delete = [Uint16LE.unpack(s) for _ in range(cIndices)]
        else:
            self.delete = []

        return self

    def __str__(self):
        return f'<CreateOffscreenBitmap {self.cx}x{self.cy} Id={self.id} Del={len(self.delete)}>'


class SwitchSurface:
    @staticmethod
    def parse(s: BytesIO) -> 'SwitchSurface':
        self = SwitchSurface()

        self.id = Uint16LE.unpack(s)
        return self

    def __str__(self):
        return f'<SwitchSurface Id={self.id}>'


class CreateNineGridBitmap:
    @staticmethod
    def parse(s: BytesIO) -> 'CreateNineGridBitmap':
        self = CreateNineGridBitmap()

        self.bpp = Uint8.unpack(s)
        self.id = Uint16LE.unpack(s)

        self.cx = Uint16LE.unpack(s)
        self.cy = Uint16LE.unpack(s)

        # NineGridInfo
        self.flFlags = Uint32LE.unpack(s)
        self.ulLeftWidth = Uint16LE.unpack(s)
        self.ulRightWidth = Uint16LE.unpack(s)
        self.ulTopHeight = Uint16LE.unpack(s)
        self.ulBottomHeight = Uint16LE.unpack(s)
        self.rgb = read_color(s)

        return self


class StreamBitmapFirst:
    @staticmethod
    def parse(s: BytesIO) -> 'StreamBitmapFirst':
        self = StreamBitmapFirst()

        self.flags = Uint8.unpack(s)
        self.bpp = Uint8.unpack(s)

        self.type = Uint16LE.unpack(s)
        self.width = Uint16LE.unpack(s)
        self.height = Uint16LE.unpack(s)

        self.totalSize = 0
        if self.flags & STREAM_BITMAP_V2:
            self.totalSize = Uint32LE.unpack(s)
        else:
            self.totalSize = Uint16LE.unpack(s)

        blockSize = Uint16LE.unpack(s)
        self.data = s.read(blockSize)

        return self


class StreamBitmapNext:
    @staticmethod
    def parse(s: BytesIO) -> 'StreamBitmapNext':
        self = StreamBitmapNext()

        self.flags = Uint8.unpack(s)
        self.bitmapType = Uint16LE.unpack(s)

        blockSize = Uint16LE.unpack(s)
        self.data = s.read(blockSize)

        return self


class GdiPlusFirst:
    @staticmethod
    def parse(s: BytesIO) -> 'GdiPlusFirst':
        self = GdiPlusFirst()

        s.read(1)  # Padding

        cbSize = Uint16LE.unpack(s)
        self.totalSize = Uint32LE.unpack(s)
        self.totalEmfSize = Uint32LE.unpack(s)
        self.data = s.read(cbSize)

        return self

    def __str__(self):
        return f'<GdiPlusFirst Size={len(self.data)} Total={self.totalSize}, EmfSize={self.totalEmfSize}>'


class GdiPlusNext:
    @staticmethod
    def parse(s: BytesIO) -> 'GdiPlusNext':
        self = GdiPlusNext()

        s.read(1)  # Padding

        cbSize = Uint16LE.unpack(s)
        self.data = s.read(cbSize)

        return self

    def __str__(self):
        return f'<GdiPlusNext Size={len(self.data)}>'


class GdiPlusEnd:
    @staticmethod
    def parse(s: BytesIO) -> 'GdiPlusEnd':
        self = GdiPlusEnd()

        s.read(1)  # Padding

        cbSize = Uint16LE.unpack(s)
        self.totalSize = Uint32LE.unpack(s)
        self.totalEmfSize = Uint32LE.unpack(s)
        self.data = s.read(cbSize)

        return self


class GdiPlusCacheFirst:
    @staticmethod
    def parse(s: BytesIO) -> 'GdiPlusCacheFirst':
        self = GdiPlusCacheFirst()

        self.flags = Uint8.unpack(s)
        self.cacheType = Uint16LE.unpack(s)
        self.cacheIdx = Uint16LE.unpack(s)

        cbSize = Uint16LE.unpack(s)
        self.totalSize = Uint32LE.unpack(s)
        self.data = s.read(cbSize)

        return self


class GdiPlusCacheNext:
    @staticmethod
    def parse(s: BytesIO) -> 'GdiPlusCacheNext':
        self = GdiPlusCacheNext()

        self.flags = Uint8.unpack(s)
        self.cacheType = Uint16LE.unpack(s)
        self.cacheIdx = Uint16LE.unpack(s)

        cbSize = Uint16LE.unpack(s)
        self.data = s.read(cbSize)

        return self


class GdiPlusCacheEnd:
    @staticmethod
    def parse(s: BytesIO) -> 'GdiPlusCacheEnd':
        self = GdiPlusCacheEnd()

        self.flags = Uint8.unpack(s)
        self.cacheType = Uint16LE.unpack(s)
        self.cacheIndex = Uint16LE.unpack(s)
        cbSize = Uint16LE.unpack(s)
        self.totalSize = Uint32LE.unpack(s)
        self.data = s.read(cbSize)

        return self


class FrameMarker:
    @staticmethod
    def parse(s: BytesIO) -> 'FrameMarker':
        self = FrameMarker()
        self.action = Uint32LE.unpack(s)

        return self

    def __str__(self):
        a = 'BEGIN' if self.action == 0 else 'END'
        return f'<Frame {a}>'

# class Window:
#     @staticmethod
#     def parse(s: BytesIO) -> 'Window':
#         self = Window()
#         # This is specified in MS-RDPERP for seamless applications.
#         return self


# class CompDeskFirst
#     @staticmethod
#     def parse(s: BytesIO) -> 'CompDeskFirst':
#         self = CompdeskFirst()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from pyrdp.core import ObservedBy
from pyrdp.enum import MCSPDUType
from pyrdp.layer.layer import Layer, LayerStrictRoutedObserver
from pyrdp.parser import MCSParser
from pyrdp.pdu import MCSAttachUserConfirmPDU, MCSAttachUserRequestPDU, MCSChannelJoinConfirmPDU, \
    MCSChannelJoinRequestPDU, MCSConnectInitialPDU, MCSConnectResponsePDU, MCSDisconnectProviderUltimatumPDU, \
    MCSDomainParams, MCSErectDomainRequestPDU, MCSSendDataIndicationPDU, MCSSendDataRequestPDU


class MCSObserver(LayerStrictRoutedObserver):
    """
    Base observer class for MCS layers. Simply routes PDU types to methods.
    """
    def __init__(self, **kwargs):
        LayerStrictRoutedObserver.__init__(self, {
            MCSPDUType.CONNECT_INITIAL: "onConnectInitial",
            MCSPDUType.CONNECT_RESPONSE: "onConnectResponse",
            MCSPDUType.ERECT_DOMAIN_REQUEST: "onErectDomainRequest",
            MCSPDUType.DISCONNECT_PROVIDER_ULTIMATUM: "onDisconnectProviderUltimatum",
            MCSPDUType.ATTACH_USER_REQUEST: "onAttachUserRequest",
            MCSPDUType.ATTACH_USER_CONFIRM: "onAttachUserConfirm",
            MCSPDUType.CHANNEL_JOIN_REQUEST: "onChannelJoinRequest",
            MCSPDUType.CHANNEL_JOIN_CONFIRM: "onChannelJoinConfirm",
            MCSPDUType.SEND_DATA_REQUEST: "onSendDataRequest",
            MCSPDUType.SEND_DATA_INDICATION: "onSendDataIndication",
        }, **kwargs)

    def onConnectInitial(self, pdu: MCSConnectInitialPDU):
        pass

    def onConnectResponse(self, pdu: MCSConnectResponsePDU):
        pass

    def onDisconnectProviderUltimatum(self, pdu: MCSDisconnectProviderUltimatumPDU):
        pass

    def onErectDomainRequest(self, pdu: MCSErectDomainRequestPDU):
        pass

    def onAttachUserRequest(self, pdu: MCSAttachUserRequestPDU):
        pass

    def onAttachUserConfirm(self, pdu: MCSAttachUserConfirmPDU):
        pass

    def onChannelJoinRequest(self, pdu: MCSChannelJoinRequestPDU):
        pass

    def onChannelJoinConfirm(self, pdu: MCSChannelJoinConfirmPDU):
        pass

    def onSendDataRequest(self, pdu: MCSSendDataRequestPDU):
        pass

    def onSendDataIndication(self, pdu: MCSSendDataIndicationPDU):
        pass


@ObservedBy(MCSObserver)
class MCSLayer(Layer):
    """
    Layer to handle MCS-related traffic.
    It doesn't really make sense to assign a single 'next' layer to this
    (since MCS is channel-based), so traffic is never forwarded.
    """

    def __init__(self, parser = MCSParser()):
        super().__init__(parser)

    def sendConnectInitial(self, payload = b"", callingDomain = b"\x01", calledDomain = b"\x01", upward = True,
                           targetParams = MCSDomainParams.createTarget(34, 2), minParams = MCSDomainParams.createMinimum(), maxParams = MCSDomainParams.createMaximum()):
        pdu = MCSConnectInitialPDU(callingDomain, calledDomain, upward, targetParams, minParams, maxParams, payload)
        self.sendPDU(pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from logging import LoggerAdapter

from pyrdp.mitm.state import RDPMITMState
from pyrdp.enum import PointerFlag, ScanCode
from pyrdp.enum.scancode import getKeyName
from pyrdp.pdu.pdu import PDU
from pyrdp.layer.layer import Layer
from pyrdp.logging.StatCounter import StatCounter, STAT

class BasePathMITM:
    """
    Base MITM component for the fast-path and slow-path layers.
    """

    def __init__(self, state: RDPMITMState, client: Layer, server: Layer, statCounter: StatCounter, log: LoggerAdapter):
        self.state = state
        self.client = client
        self.server = server
        self.statCounter = statCounter
        self.log = log

    def onClientPDUReceived(self, pdu: PDU):
        raise NotImplementedError("onClientPDUReceived must be overridden")

    def onServerPDUReceived(self, pdu: PDU):
        raise NotImplementedError("onServerPDUReceived must be overridden")

    def loginAttempt(self):
        if self.state.loggedIn or self.state.inputBuffer == "":
            return

        self.state.credentialsCandidate = self.state.inputBuffer
        self.state.inputBuffer = ""

        self.log.info("Credentials attempt from heuristic: %(credentials_attempt)s", {
            "credentials_attempt": (self.state.credentialsCandidate)
        })

    def onMouse(self, mouseX: int, mouseY: int, pointerFlags: int):
        if pointerFlags & PointerFlag.PTRFLAGS_DOWN != 0:
            percentageX = mouseX / self.state.windowSize[0]
            percentageY = mouseY / self.state.windowSize[1]

            if 0.5 < percentageX < 0.65 and 0.5 < percentageY < 0.65:
                self.loginAttempt()

    def onScanCode(self, scanCode: int, isReleased: bool, isExtended: bool):
        """
        Handle scan code.
        """
        keyName = getKeyName(scanCode, isExtended, self.state.shiftPressed, self.state.capsLockOn)
        scanCodeTuple = (scanCode, isExtended)

        # Left or right shift
        if scanCodeTuple in [ScanCode.LSHIFT, ScanCode.RSHIFT]:
            self.state.shiftPressed = not isReleased
        # Caps lock
        elif scanCodeTuple == ScanCode.CAPSLOCK and not isReleased:
            self.state.capsLockOn = not self.state.capsLockOn
        # Control
        elif scanCodeTuple in [ScanCode.LCONTROL, ScanCode.RCONTROL]:
            self.state.ctrlPressed = not isReleased
        # Backspace
        elif scanCodeTuple == ScanCode.BACKSPACE and not isReleased:
            self.state.inputBuffer += "<\\b>"
        # Tab
        elif scanCodeTuple == ScanCode.TAB and not isReleased:
            self.state.inputBuffer += "<\\t>"
        # CTRL + A
        elif scanCodeTuple == ScanCode.KEY_A and self.state.ctrlPressed and not isReleased:
            self.state.inputBuffer += "<ctrl-a>"
        elif scanCodeTuple == ScanCode.SPACE and not isReleased:
            self.state.inputBuffer += " "
        # Return
        elif scanCodeTuple == ScanCode.RETURN and not isReleased:
            self.loginAttempt()
        # Normal input
        elif len(keyName) == 1:
            if not isReleased:
                self.state.inputBuffer += keyName



#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
import operator
from typing import Callable, Dict, List


class Event:
    """
    An asyncio.Event wrapper that will be triggered when an object passes a check.
    """
    def __init__(self):
        self.aioEvent = asyncio.Event()
        self.object = None

    def check(self, obj) -> bool:
        """
        Check if an object is a match for this event.
        :param obj: the object we are trying to match.
        :return: True if the object matches, otherwise False.
        """
        raise NotImplementedError("check must be overridden")


class FunctionEvent(Event):
    """
    An asyncio.Event wrapper that will be triggered when a given callable returns True.
    """
    def __init__(self, predicate: Callable[[object], bool]):
        super().__init__()
        self.predicate = predicate

    def check(self, obj) -> bool:
        """
        Check if an object is a match for this event by calling the provided callable.
        :param obj: the object we are trying to match.
        :return: True if the object matches, otherwise False.
        """
        return self.predicate(obj)


class PropertyEvent(Event):
    """
    An asyncio.Event wrapper that will be triggered when an object's attributes have specific values.
    """
    def __init__(self, attributes: Dict[str, object]):
        super().__init__()
        self.attributes = {operator.attrgetter(key): value for key, value in attributes.items()}

    def check(self, obj) -> bool:
        """
        Check if an object is a match for this event by checking that its attributes match the given dict.
        :param obj: the object we are trying to match.
        :return: True if the object matches, otherwise False.
        """
        for getter, expectedValue in self.attributes.items():
            try:
                attributeValue = getter(obj)
            except AttributeError:
                return False

            if attributeValue != expectedValue:
                return False

        return True


class EventEngine:
    """
    An event engine is a queue of asyncio.Event wrappers.
    Every time an object of interest is received, it should be passed to processObject.
    The event engine will try to match the object with every event being awaited until a match is made (if any).
    If a match is made, the event that was checked will be triggered and the object will be consumed (only one match is allowed per object).
    If an object could potentially be matched with multiple events, only the event that was added first will be triggered.

    Users of this class are expected to call the wait function and pass it attributes that it should have and / or a predicate
    function that will be used to check if the objects that are received match. Example:

    class Foo:
        def __init__(self, value):
            self.attribute = value

    class Bar:
        attribute = 42

    engine = EventEngine()

    # ...

    async def myCoroutine():
        foo = await engine.wait({"attribute": 42})
        print(foo.attribute) # Prints 42

    # ...

    # This will resolve the await in myCoroutine
    engine.processObject(Foo(42))

    # This will not resolve the await in myCoroutine because attribute != 42
    engine.processObject(Foo(123))

    # This will also resolve because it has an attribute called "attribute" with a value of 42
    engine.processObject(Bar(42))
    """

    @staticmethod
    def Anything(_):
        """
        This predicate can be used to wait for any object.
        """
        return True

    def __init__(self):
        """
        Create a new (empty) event engine.
        """
        self.events: List[Event] = []

    def processObject(self, obj):
        """
        Call when an object of interest is received. The engine will try to match the object to events currently awaited.
        :param obj: the object of interest.
        :return: True if a match was made, otherwise False.
        """
        for event in self.events:
            if event.check(obj):
                self.events.remove(event)
                event.object = obj
                event.aioEvent.set()
                return True

        return False

    async def wait(self, where: Dict[str, object] = None, match: Callable[[object], bool] = None):
        """
        Call to wait for an object matching certain criteria to arrive.
        If both predicate and attributes are not None, then attributes will be matched first, then predicate.
        If both predicate and attributes are None, then any object will be matched.
        :param where: dict of attribute name -> value that the object should have.
        :param match: function that will be used to match the object (returns True for matching objects).
        :return: the object that matched the criteria.
        """
        propertyEvent: PropertyEvent = None

        if where:
            propertyEvent = PropertyEvent(where)

        if match is not None:
            if propertyEvent:
                event = FunctionEvent(lambda obj: propertyEvent.check(obj) and match(obj))
            else:
                event = FunctionEvent(match)
        elif propertyEvent:
            event = propertyEvent
        else:
            event = FunctionEvent(EventEngine.Anything)

        self.events.append(event)
        await event.aioEvent.wait()
        return event.object



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum


class ParserMode(IntEnum):
    """
    Mode used by some parsers (Client or Server).
    """
    CLIENT = 0
    SERVER = 1


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from pyrdp.enum import VirtualChannelPDUFlag
from pyrdp.pdu.pdu import PDU


class VirtualChannelPDU(PDU):
    """
    https://msdn.microsoft.com/en-us/library/cc240553.aspx
    """

    def __init__(self, flags: VirtualChannelPDUFlag, payload: bytes):
        """
        :param flags: PDU flags (CHANNEL_FLAG_FIRST and CHANNEL_FLAG_LAST are added automatically when writing).
        :param payload: the payload.
        """
        PDU.__init__(self, payload)
        self.flags = flags



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum
from pathlib import PosixPath
from typing import List

from pyrdp.core import ObservedBy, Observer, Subject


class FileSystemItemType(IntEnum):
    Directory = 1
    Drive = 2
    File = 3


class FileSystemItem:
    def __init__(self, name: str, itemType: FileSystemItemType):
        super().__init__()
        self.name = name
        self.type = itemType

    def getFullPath(self, name: str = "") -> str:
        pass


class DirectoryObserver(Observer):
    def onDirectoryChanged(self):
        pass

    def onListDirectory(self, deviceID: int, path: str):
        pass

@ObservedBy(DirectoryObserver)
class Directory(FileSystemItem, Subject):
    def __init__(self, name: str, parent: 'Directory' = None):
        super().__init__(name, FileSystemItemType.Directory)
        self.parent = parent
        self.files: List[File] = []
        self.directories: List[Directory] = []

    def addFile(self, name: str):
        file = File(name, self)
        self.files.append(file)

        self.observer.onDirectoryChanged()

        return file

    def addDirectory(self, name: str):
        directory = Directory(name, self)
        self.directories.append(directory)

        self.observer.onDirectoryChanged()

        return directory

    def list(self, name: str = ""):
        if name == "":
            self.files.clear()
            self.directories.clear()

        path = self.getFullPath(name)
        self.parent.list(str(path))

    def getFullPath(self, name: str = "") -> str:
        path = PosixPath(self.name)

        if name != "":
            path /= name

        path = str(path)

        if self.parent is None:
            return path
        else:
            return self.parent.getFullPath(path)

    def getRootParent(self) -> FileSystemItem:
        parent = self.parent

        while parent.parent is not None:
            parent = parent.parent

        return parent

class File(FileSystemItem):
    def __init__(self, name: str, parent: Directory):
        super().__init__(name, FileSystemItemType.File)
        self.parent = parent

    def getFullPath(self, name: str = "") -> str:
        path = PosixPath(self.name)

        if name != "":
            path /= name

        path = str(path)

        if self.parent is None:
            return path
        else:
            return self.parent.getFullPath(path)

    def getRootParent(self) -> FileSystemItem:
        parent = self.parent

        while parent.parent is not None:
            parent = parent.parent

        return parent

class Drive(Directory):
    def __init__(self, name: str, deviceID: int):
        super().__init__(name, None)
        self.type = FileSystemItemType.Drive
        self.deviceID = deviceID

    def list(self, name: str = ""):
        path = "/"

        if name != "":
            path += name
        else:
            self.files.clear()
            self.directories.clear()

        self.observer.onListDirectory(self.deviceID, path)

    def getFullPath(self, name: str = "") -> str:
        path = PosixPath("/")

        if name != "":
            path /= name

        return str(path)

@ObservedBy(DirectoryObserver)
class FileSystem(Directory):
    def __init__(self):
        super().__init__("")

    def addDrive(self, name: str, deviceID: int) -> Drive:
        drive = Drive(name, deviceID)
        self.directories.append(drive)

        self.observer.onDirectoryChanged()

        return drive

    def list(self, name: str = ""):
        pass


#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from typing import List, Optional, Union

from pyrdp.enum import SegmentationPDUType
from pyrdp.pdu.pdu import PDU
from pyrdp.pdu.rdp.bitmap import BitmapUpdateData
from pyrdp.pdu.segmentation import SegmentationPDU


class FastPathEvent(PDU):
    """
    Base class for RDP fast path event (not PDU, a PDU contains multiple events).
    Used for scan code events, mouse events or bitmap events.
    """

    def __init__(self, payload: bytes = b""):
        super().__init__(payload)


class FastPathPDU(SegmentationPDU):
    def __init__(self, header: int, events: List[FastPathEvent]):
        super().__init__(b"")
        self.header = header
        self.events = events

    def getSegmentationType(self) -> SegmentationPDUType:
        return SegmentationPDUType.FAST_PATH

    def __repr__(self) -> str:
        return str([str(e.__class__) for e in self.events])


class FastPathEventRaw(FastPathEvent):
    def __init__(self, data: bytes):
        super().__init__()
        self.data = data


class FastPathInputEvent(FastPathEvent):
    def __init__(self):
        super().__init__()


class FastPathOutputEvent(FastPathEvent):
    """
    https://msdn.microsoft.com/en-us/library/cc240622.aspx
    """

    def __init__(self, header: int, compressionFlags: Optional[int], payload: bytes = b""):
        super().__init__(payload)
        self.header = header
        self.compressionFlags = compressionFlags


class FastPathScanCodeEvent(FastPathInputEvent):

    def __init__(self, rawHeaderByte: int, scanCode: int, isReleased: bool):
        super().__init__()
        self.rawHeaderByte = rawHeaderByte
        self.scanCode = scanCode
        self.isReleased = isReleased


class FastPathMouseEvent(FastPathInputEvent):
    """
    Mouse event (clicks, move, scroll, etc.)
    """

    def __init__(self, rawHeaderByte: int, pointerFlags: int, mouseX: int, mouseY: int):
        super().__init__()
        self.rawHeaderByte = rawHeaderByte
        self.mouseY = mouseY
        self.mouseX = mouseX
        self.pointerFlags = pointerFlags


class FastPathUnicodeEvent(FastPathInputEvent):
    """
    Unicode event (text presses and releases)
    """

    def __init__(self, text: Union[str, bytes], released: bool):
        super().__init__()
        self.text = text
        self.released = released


class FastPathBitmapEvent(FastPathOutputEvent):
    def __init__(self, header: int, compressionFlags: Optional[int],
                 bitmapUpdateData: List[BitmapUpdateData], payload: bytes):
        super().__init__(header, compressionFlags, payload)
        self.bitmapUpdateData = bitmapUpdateData


class FastPathOrdersEvent(FastPathOutputEvent):
    """
    Encapsulate drawing orders.

    https://msdn.microsoft.com/en-us/library/cc241573.aspx
    """

    def __init__(self, header: int, compressionFlags: Optional[int], payload: bytes):
        super().__init__(header, compressionFlags, payload=payload)
        self.compressionFlags = compressionFlags



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.core import Uint16LE, Uint32LE
from pyrdp.enum import PointerEventType
from pyrdp.exceptions import ParsingError
from pyrdp.parser.parser import Parser, StreamParser
from pyrdp.pdu import Point, PointerCacheEvent, PointerColorEvent, PointerNewEvent, PointerPositionEvent, \
    PointerSystemEvent


class PointerEventParser(StreamParser):
    def __init__(self):
        super().__init__()
        self.parsers = {
            PointerEventType.TS_PTRMSGTYPE_SYSTEM: self.parseSystemEvent,
            PointerEventType.TS_PTRMSGTYPE_POSITION: self.parsePositionEvent,
            PointerEventType.TS_PTRMSGTYPE_COLOR: self.parseColorEvent,
            PointerEventType.TS_PTRMSGTYPE_CACHED: self.parseCacheEvent,
            PointerEventType.TS_PTRMSGTYPE_POINTER: self.parseNewEvent,
        }

        self.writers = {
            PointerEventType.TS_PTRMSGTYPE_SYSTEM: self.writeSystemEvent,
            PointerEventType.TS_PTRMSGTYPE_POSITION: self.writePositionevent,
            PointerEventType.TS_PTRMSGTYPE_COLOR: self.writeColorEvent,
            PointerEventType.TS_PTRMSGTYPE_CACHED: self.writeCacheEvent,
            PointerEventType.TS_PTRMSGTYPE_POINTER: self.writeNewEvent,
        }

    def doParse(self, stream):
        messageType = Uint16LE.unpack(stream)
        stream.read(2)

        if messageType not in self.parsers:
            raise ParsingError("Trying to parse invalid pointer event type")

        return self.parsers[messageType](stream)

    def write(self, event):
        stream = BytesIO()
        Uint16LE.pack(event.messageType, stream)
        stream.write(b"\x00" * 2)

        if event.messageType not in self.writers:
            raise ParsingError("Trying to write invalid pointer event type")

        self.writers[event.messageType](stream, event)
        return stream.getvalue()

    def parseSystemEvent(self, stream):
        pointerType = Uint32LE.unpack(stream)
        return PointerSystemEvent(pointerType)

    def writeSystemEvent(self, stream, event):
        Uint32LE.pack(event.pointerType, stream)

    def parsePositionEvent(self, stream):
        x = Uint16LE.unpack(stream)
        y = Uint16LE.unpack(stream)
        return PointerPositionEvent(Point(x, y))

    def writePositionevent(self, stream, event):
        Uint16LE.pack(event.point.x, stream)
        Uint16LE.pack(event.point.y, stream)

    def parseColorEvent(self, stream):
        cacheIndex = Uint16LE.unpack(stream)
        hotSpot = Uint32LE.unpack(stream)
        width = Uint16LE.unpack(stream)
        height = Uint16LE.unpack(stream)
        andMaskLength = Uint16LE.unpack(stream)
        xorMaskLength = Uint16LE.unpack(stream)
        xorMask = stream.read(xorMaskLength)
        andMask = stream.read(andMaskLength)
        stream.read(1)

        return PointerColorEvent(cacheIndex, hotSpot, width, height, andMask, xorMask)

    def writeColorEvent(self, stream, event):
        Uint16LE.pack(event.cacheIndex, stream)
        Uint32LE.pack(event.hotSpot, stream)
        Uint16LE.pack(event.width, stream)
        Uint16LE.pack(event.height, stream)
        Uint16LE.pack(len(event.andMask), stream)
        Uint16LE.pack(len(event.xorMask), stream)
        stream.write(event.xorMask)
        stream.write(event.andMask)
        stream.write(b"\x00")

    def parseCacheEvent(self, stream):
        cacheIndex = Uint16LE.unpack(stream)
        return PointerCacheEvent(cacheIndex)

    def writeCacheEvent(self, stream, event):
        Uint16LE.pack(event.cacheIndex, stream)

    def parseNewEvent(self, stream):
        xorBPP = Uint16LE.unpack(stream)
        color = self.parseColorEvent(stream)
        return PointerNewEvent(xorBPP, color)

    def writeNewEvent(self, stream, event):
        Uint16LE.pack(event.xorBPP, stream)
        self.writeColorEvent(stream, event.color)


#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum, IntFlag


class DeviceRedirectionComponent(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241324.aspx
    """
    RDPDR_CTYP_CORE = 0x4472
    RDPDR_CTYP_PRN = 0x5052


class DeviceRedirectionPacketID(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241324.aspx
    """
    PAKID_CORE_SERVER_ANNOUNCE = 0x496E
    PAKID_CORE_CLIENTID_CONFIRM = 0x4343
    PAKID_CORE_CLIENT_NAME = 0x434E
    PAKID_CORE_DEVICELIST_ANNOUNCE = 0x4441
    PAKID_CORE_DEVICE_REPLY = 0x6472
    PAKID_CORE_DEVICE_IOREQUEST = 0x4952
    PAKID_CORE_DEVICE_IOCOMPLETION = 0x4943
    PAKID_CORE_SERVER_CAPABILITY = 0x5350
    PAKID_CORE_CLIENT_CAPABILITY = 0x4350
    PAKID_CORE_DEVICELIST_REMOVE = 0x444D
    PAKID_PRN_CACHE_DATA = 0x5043
    PAKID_CORE_USER_LOGGEDON = 0x554C
    PAKID_PRN_USING_XPS = 0x5543


class MajorFunction(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241327.aspx
    """

    IRP_MJ_CREATE = 0x00000000
    IRP_MJ_CLOSE = 0x00000002
    IRP_MJ_READ = 0x00000003
    IRP_MJ_WRITE = 0x00000004
    IRP_MJ_DEVICE_CONTROL = 0x0000000E
    IRP_MJ_QUERY_VOLUME_INFORMATION = 0x0000000A
    IRP_MJ_SET_VOLUME_INFORMATION = 0x0000000B
    IRP_MJ_QUERY_INFORMATION = 0x00000005
    IRP_MJ_SET_INFORMATION = 0x00000006
    IRP_MJ_DIRECTORY_CONTROL = 0x0000000C
    IRP_MJ_LOCK_CONTROL = 0x00000011


class MinorFunction(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241327.aspx
    """
    IRP_MN_QUERY_DIRECTORY = 0x00000001
    IRP_MN_NOTIFY_CHANGE_DIRECTORY = 0x00000002


class CreateOption(IntEnum):
    FILE_DIRECTORY_FILE = 0x00000001
    FILE_WRITE_THROUGH = 0x00000002
    FILE_SEQUENTIAL_ONLY = 0x00000004
    FILE_NO_INTERMEDIATE_BUFFERING = 0x00000008
    FILE_SYNCHRONOUS_IO_ALERT = 0x00000010
    FILE_SYNCHRONOUS_IO_NONALERT = 0x00000020
    FILE_NON_DIRECTORY_FILE = 0x00000040
    FILE_COMPLETE_IF_OPLOCKED = 0x00000100
    FILE_NO_EA_KNOWLEDGE = 0x00000200
    FILE_RANDOM_ACCESS = 0x00000800
    FILE_DELETE_ON_CLOSE = 0x00001000
    FILE_OPEN_BY_FILE_ID = 0x00002000
    FILE_OPEN_FOR_BACKUP_INTENT = 0x00004000
    FILE_NO_COMPRESSION = 0x00008000
    FILE_OPEN_REMOTE_INSTANCE = 0x00000400
    FILE_OPEN_REQUIRING_OPLOCK = 0x00010000
    FILE_DISALLOW_EXCLUSIVE = 0x00020000
    FILE_RESERVE_OPFILTER = 0x00100000
    FILE_OPEN_REPARSE_POINT = 0x00200000
    FILE_OPEN_NO_RECALL = 0x00400000
    FILE_OPEN_FOR_FREE_SPACE_QUERY = 0x00800000


class DeviceType(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241326.aspx
    """

    RDPDR_DTYP_SERIAL = 0x00000001
    RDPDR_DTYP_PARALLEL = 0x00000002
    RDPDR_DTYP_PRINT = 0x00000004
    RDPDR_DTYP_FILESYSTEM = 0x00000008
    RDPDR_DTYP_SMARTCARD = 0x00000020

    @staticmethod
    def getPrettyName(deviceType: 'DeviceType'):
        if deviceType == DeviceType.RDPDR_DTYP_FILESYSTEM:
            return "Filesystem"
        elif deviceType == DeviceType.RDPDR_DTYP_SMARTCARD:
            return "Smart card"
        elif deviceType == DeviceType.RDPDR_DTYP_PRINT:
            return "Printer"
        elif deviceType == DeviceType.RDPDR_DTYP_PARALLEL:
            return "Parallel port"
        elif deviceType == DeviceType.RDPDR_DTYP_SERIAL:
            return "Serial port"
        else:
            return str(deviceType)


class RDPDRCapabilityType(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241325.aspx
    """
    CAP_GENERAL_TYPE = 0x0001
    CAP_PRINTER_TYPE = 0x0002
    CAP_PORT_TYPE = 0x0003
    CAP_DRIVE_TYPE = 0x0004
    CAP_SMARTCARD_TYPE = 0x0005


class GeneralCapabilityVersion(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241349.aspx
    """
    GENERAL_CAPABILITY_VERSION_01 = 0x00000001
    GENERAL_CAPABILITY_VERSION_02 = 0x00000002


class ExtendedPDUFlags(IntEnum):
    """
    https://msdn.microsoft.com/en-us/library/cc241349.aspx
    """
    RDPDR_DEVICE_REMOVE_PDUS = 0x00000001
    RDPDR_CLIENT_DISPLAY_NAME_PDU = 0x00000002
    RDPDR_USER_LOGGEDON_PDU = 0x00000004


class FileSystemInformationClass(IntEnum):
    FileDirectoryInformation = 0x00000001
    FileFullDirectoryInformation = 0x00000002
    FileBothDirectoryInformation = 0x00000003
    FileNamesInformation = 0x0000000C


class FileAttributes(IntFlag):
    FILE_ATTRIBUTE_NONE = 0x00000000
    FILE_ATTRIBUTE_ARCHIVE = 0x00000020
    FILE_ATTRIBUTE_COMPRESSED = 0x00000800
    FILE_ATTRIBUTE_DIRECTORY = 0x00000010
    FILE_ATTRIBUTE_ENCRYPTED = 0x00004000
    FILE_ATTRIBUTE_HIDDEN = 0x00000002
    FILE_ATTRIBUTE_NORMAL = 0x00000080
    FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = 0x00002000
    FILE_ATTRIBUTE_OFFLINE = 0x00001000
    FILE_ATTRIBUTE_READONLY = 0x00000001
    FILE_ATTRIBUTE_REPARSE_POINT = 0x00000400
    FILE_ATTRIBUTE_SPARSE_FILE = 0x00000200
    FILE_ATTRIBUTE_SYSTEM = 0x00000004
    FILE_ATTRIBUTE_TEMPORARY = 0x00000100
    FILE_ATTRIBUTE_INTEGRITY_STREAM = 0x00008000
    FILE_ATTRIBUTE_NO_SCRUB_DATA = 0x00020000


class FileAccessMask(IntFlag):
    """
    https://msdn.microsoft.com/en-us/library/cc246802.aspx
    """

    FILE_READ_DATA = 0x00000001
    FILE_WRITE_DATA = 0x00000002
    FILE_APPEND_DATA = 0x00000004
    FILE_READ_EA = 0x00000008
    FILE_WRITE_EA = 0x00000010
    FILE_DELETE_CHILD = 0x00000040
    FILE_EXECUTE = 0x00000020
    FILE_READ_ATTRIBUTES = 0x00000080
    FILE_WRITE_ATTRIBUTES = 0x00000100
    DELETE = 0x00010000
    READ_CONTROL = 0x00020000
    WRITE_DAC = 0x00040000
    WRITE_OWNER = 0x00080000
    SYNCHRONIZE = 0x00100000
    ACCESS_SYSTEM_SECURITY = 0x01000000
    MAXIMUM_ALLOWED = 0x02000000
    GENERIC_ALL = 0x10000000
    GENERIC_EXECUTE = 0x20000000
    GENERIC_WRITE = 0x40000000
    GENERIC_READ = 0x80000000


class DirectoryAccessMask(IntFlag):
    FILE_LIST_DIRECTORY = 0x00000001
    FILE_ADD_FILE = 0x00000002
    FILE_ADD_SUBDIRECTORY = 0x00000004
    FILE_READ_EA = 0x00000008
    FILE_WRITE_EA = 0x00000010
    FILE_TRAVERSE = 0x00000020
    FILE_DELETE_CHILD = 0x00000040
    FILE_READ_ATTRIBUTES = 0x00000080
    FILE_WRITE_ATTRIBUTES = 0x00000100
    DELETE = 0x00010000
    READ_CONTROL = 0x00020000
    WRITE_DAC = 0x00040000
    WRITE_OWNER = 0x00080000
    SYNCHRONIZE = 0x00100000
    ACCESS_SYSTEM_SECURITY = 0x01000000
    MAXIMUM_ALLOWED = 0x02000000
    GENERIC_ALL = 0x10000000
    GENERIC_EXECUTE = 0x20000000
    GENERIC_WRITE = 0x40000000
    GENERIC_READ = 0x80000000


class FileShareAccess(IntFlag):
    FILE_SHARE_READ = 0x00000001
    FILE_SHARE_WRITE = 0x00000002
    FILE_SHARE_DELETE = 0x00000004


class FileCreateDisposition(IntEnum):
    FILE_SUPERSEDE = 0x00000000
    FILE_OPEN = 0x00000001
    FILE_CREATE = 0x00000002
    FILE_OPEN_IF = 0x00000003
    FILE_OVERWRITE = 0x00000004
    FILE_OVERWRITE_IF = 0x00000005


class FileCreateOptions(IntFlag):
    FILE_DIRECTORY_FILE = 0x00000001
    FILE_WRITE_THROUGH = 0x00000002
    FILE_SEQUENTIAL_ONLY = 0x00000004
    FILE_NO_INTERMEDIATE_BUFFERING = 0x00000008
    FILE_SYNCHRONOUS_IO_ALERT = 0x00000010
    FILE_SYNCHRONOUS_IO_NONALERT = 0x00000020
    FILE_NON_DIRECTORY_FILE = 0x00000040
    FILE_COMPLETE_IF_OPLOCKED = 0x00000100
    FILE_NO_EA_KNOWLEDGE = 0x00000200
    FILE_RANDOM_ACCESS = 0x00000800
    FILE_DELETE_ON_CLOSE = 0x00001000
    FILE_OPEN_BY_FILE_ID = 0x00002000
    FILE_OPEN_FOR_BACKUP_INTENT = 0x00004000
    FILE_NO_COMPRESSION = 0x00008000
    FILE_OPEN_REMOTE_INSTANCE = 0x00000400
    FILE_OPEN_REQUIRING_OPLOCK = 0x00010000
    FILE_DISALLOW_EXCLUSIVE = 0x00020000
    FILE_RESERVE_OPFILTER = 0x00100000
    FILE_OPEN_REPARSE_POINT = 0x00200000
    FILE_OPEN_NO_RECALL = 0x00400000
    FILE_OPEN_FOR_FREE_SPACE_QUERY = 0x00800000


#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
"""
Drawing Order Context.
"""

from pyrdp.parser.rdp.orders.alternate import CreateOffscreenBitmap, SwitchSurface, CreateNineGridBitmap, \
    StreamBitmapFirst, StreamBitmapNext, GdiPlusFirst, GdiPlusNext, GdiPlusEnd, GdiPlusCacheFirst, \
    GdiPlusCacheNext, GdiPlusCacheEnd, FrameMarker

from .common import Bounds
from pyrdp.enum import IntEnum


class BrushStyle(IntEnum):
    SOLID = 0x00
    NULL = 0x01
    HATCHED = 0x02
    PATTERN = 0x03


class HatchStyle(IntEnum):
    HORIZONTAL = 0x00
    VERTICAL = 0x01
    FDIAGONAL = 0x02
    BDIAGNOAL = 0x03
    CROSS = 0x04
    DIAGCROSS = 0x05


class GdiFrontend:
    """
    Interface for objects that implement GDI.

    This class provides abstract methods to be used by modules
    interested in listening and acting upon context updates.
    Its primary purpose is for the PyRDP player to render the
    remote desktop.

    NOTE: Unimplemented methods will act as No-Op.
    """
    # REFACTOR: Move to core, this isn't really relevant to the parser.

    def onBounds(self, b: Bounds):
        """
        Called by the parser to configure the bounding rectangle.
        """
        pass

    def dstBlt(self, state):
        pass

    def patBlt(self, state):
        pass

    def scrBlt(self, state):
        pass

    def drawNineGrid(self, state):
        pass

    def multiDrawNineGrid(self, state):
        pass

    def lineTo(self, state):
        pass

    def opaqueRect(self, state):
        pass

    def saveBitmap(self, state):
        pass

    def memBlt(self, state):
        pass

    def mem3Blt(self, state):
        pass

    def multiDstBlt(self, state):
        pass

    def multiPatBlt(self, state):
        pass

    def multiScrBlt(self, state):
        pass

    def multiOpaqueRect(self, state):
        pass

    def fastIndex(self, state):
        pass

    def polygonSc(self, state):
        pass

    def polygonCb(self, state):
        pass

    def polyLine(self, state):
        pass

    def fastGlyph(self, state):
        pass

    def ellipseSc(self, state):
        pass

    def ellipseCb(self, state):
        pass

    def glyphIndex(self, state):
        pass

    # Secondary Handlers
    def cacheBitmapV1(self, state):
        pass

    def cacheBitmapV2(self, state):
        pass

    def cacheBitmapV3(self, state):
        pass

    def cacheColorTable(self, state):
        pass

    def cacheGlyph(self, state):
        pass

    def cacheBrush(self, state):
        pass

    # Alternate Secondary Handlers
    def frameMarker(self, state: FrameMarker):
        pass

    def createOffscreenBitmap(self, state: CreateOffscreenBitmap):
        """
        Create an offscreen bitmap.
        """
        pass

    def switchSurface(self, state: SwitchSurface):
        """
        Switch drawing surface.
        """
        pass

    def createNineGridBitmap(self, state: CreateNineGridBitmap):
        """
        Create a Nine Grid bitmap.
        """
        pass

    def streamBitmapFirst(self, state: StreamBitmapFirst):
        pass

    def streamBitmapNext(self, state: StreamBitmapNext):
        pass

    def drawGdiPlusFirst(self, state: GdiPlusFirst):
        pass

    def drawGdiPlusNext(self, state: GdiPlusNext):
        pass

    def drawGdiPlusEnd(self, state: GdiPlusEnd):
        pass

    def drawGdiPlusCacheFirst(self, state: GdiPlusCacheFirst):
        pass

    def drawGdiPlusCacheNext(self, state: GdiPlusCacheNext):
        pass

    def drawGdiPlusCacheEnd(self, state: GdiPlusCacheEnd):
        pass
